from typing import Optional

import numpy as np
import pandas as pd

from relbench.base import Database, Table

from ctu_relational.utils import TIMESTAMP_MIN, TIMESTAMP_MAX

from .ctu_dataset import CTUDataset

# fmt: off
__all__ = [    
    "Accidents", "AdventureWorks", "Airline", "Atherosclerosis", "BasketballMen",
    "BasketballWomen", "Biodegradability", "Bupa", "Carcinogenesis",
    "CDESchools", "Chess", "ClassicModels", "CORA", "Countries", "CraftBeer", "Credit",
    "Dallas", "DCG", "Diabetes", "Dunur", "Elti", "Employee", "ErgastF1",
    "Expenditures", "Employee", "Financial", "FNHK", "FTP", "Geneea", "Genes", "GOSales",
    "Grants", "Hepatitis", "Hockey", "IMDb", "Lahman", "LegalActs", "Mesh",
    "Mondial", "Mooney", "MovieLens", "MuskLarge", "MuskSmall", "Mutagenesis",
    "NCAA", "Northwind", "Pima", "PremiereLeague", "Restbase", "Sakila",
    "Sales", "SameGen", "SAP", "Satellite", "Seznam", "SFScores", "Shakespeare", "Stats",
    "StudentLoan", "Thrombosis", "Toxicology", "TPCC", "TPCD", "TPCDS", "TPCH", "Triazine",
    "UWCSE", "VisualGenome", "VOC", "Walmart", "WebKP", "World"
]
# fmt: on


class Accidents(CTUDataset):
    """
    Traffic accident database consists of all accidents that happened in Slovenia's\
    capital city Ljubljana between the years 1995 and 2006.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Accidents",
            cache_dir=cache_dir,
            time_col_dict={"nesreca": "cas_nesreca"},
            keep_original_keys=False,
        )


class AdventureWorks(CTUDataset):
    """
    Adventure Works 2014 (OLTP version) is a sample database for Microsoft SQL Server, \
    which has replaced Northwind and Pub sample databases that were shipped earlier. \
    The database is about a fictious, multinational bicycle manufacturer called \
    Adventure Works Cycles.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "AdventureWorks2014",
            cache_dir=cache_dir,
            time_col_dict={
                # "BillOfMaterials": "StartDate",
                # "Employee": "HireDate",
                # "EmployeePayHistory": "RateChangeDate",
                # "CurrencyRate": "CurrencyRateDate",
                # "Product": "SellStartDate",
                # "ProductCostHistory": "StartDate",
                # "ProductListPriceHistory": "StartDate",
                # "ProductReview": "ReviewDate",
                # "PurchaseOrderHeader": "OrderDate",
                "SalesOrderHeader": "OrderDate",
                # "SalesPersonQuotaHistory": "QuotaDate",
                # "SalesTerritoryHistory": "StartDate",
                # "SpecialOffer": "StartDate",
                # "TransactionHistory": "TransactionDate",
                # "WorkOrder": "StartDate",
                # "WorkOrderRouting": "ScheduledStartDate",
                # "EmployeeDepartmentHistory": "StartDate",
            },
            keep_original_keys=False,
            keep_original_compound_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict.pop("AWBuildVersion", None)
        db.table_dict.pop("DatabaseLog", None)
        db.table_dict.pop("ErrorLog", None)
        db.table_dict.pop("TransactionHistoryArchive", None)

        db.table_dict["Address"].df.drop(columns=["SpatialLocation"], inplace=True)
        db.table_dict["Document"].df.drop(columns=["Document"], inplace=True)
        db.table_dict["ProductPhoto"].df.drop(columns=["ThumbNailPhoto"], inplace=True)
        db.table_dict["ProductPhoto"].df.drop(columns=["LargePhoto"], inplace=True)

        return db


class Airline(CTUDataset):
    """
    Airline on-time data are reported each month to the U.S. Department of Transportation\
    (DOT), Bureau of Transportation Statistics (BTS) by the 16 U.S. air carriers that have\
    at least 1 percent of total domestic scheduled-service passenger revenues, plus two\
    other carriers that report voluntarily. The data cover nonstop scheduled-service\
    flights between points within the United States (including territories) as described\
    in 14 CFR Part 234 of DOT's regulations. Data are available since January 1995.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Airline",
            cache_dir=cache_dir,
            time_col_dict={"On_Time_On_Time_Performance_2016_1": "FlightDate"},
            keep_original_keys=False,
        )


class Atherosclerosis(CTUDataset):
    """
    The study STULONG is a longitudinal 20 years lasting primary preventive study of\
    middle-aged men. The study aims to identify prevalence of atherosclerosis RFs in\
    a population generally considered to be the most endangered by possible atherosclerosis\
    complications, i.e., middle-aged men.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Atherosclerosis",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class BasketballMen(CTUDataset):
    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Basketball_men",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict["players"].df["birthDate"] = pd.to_datetime(
            db.table_dict["players"].df["birthDate"], errors="coerce"
        )
        db.table_dict["players"].df["deathDate"] = pd.to_datetime(
            db.table_dict["players"].df["deathDate"], errors="coerce"
        )

        return db


class BasketballWomen(CTUDataset):
    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Basketball_women",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Biodegradability(CTUDataset):
    """
    This is an older data set of chemical structures containing 328 compounds\
    labeled by their half-life for aerobic aqueous biodegradation.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Biodegradability",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Bupa(CTUDataset):
    """Evaluation of patients on liver disorder."""

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Bupa",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )


class Carcinogenesis(CTUDataset):
    """
    For prediction of whether a given molecule is carcinogenic or not.\
    The dataset contains 182 positive carcinogenicity tests and 148 negative tests.
    """

    ""

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Carcinogenesis",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class CDESchools(CTUDataset):
    """
    A database containing geospatial information, as well as SAT average scores and\
    Free-or-Reduced-Price Meal eligibility data, for California schools.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "CDESchools",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Chess(CTUDataset):
    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Chess",
            cache_dir=cache_dir,
            time_col_dict={"game": "event_date"},
            keep_original_keys=False,
        )


class ClassicModels(CTUDataset):
    """
    The schema is for Classic Models, a retailer of scale models of classic cars.\
    The database contains typical business data such as customers, orders, order\
    line items, products and so on.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "classicmodels",
            cache_dir=cache_dir,
            time_col_dict={"orders": "orderDate", "payments": "paymentDate"},
            keep_original_keys=False,
        )


class CORA(CTUDataset):
    """
    The Cora dataset consists of 2708 scientific publications classified into one of\
    seven classes. The citation network consists of 5429 links. Each publication in the\
    dataset is described by a 0/1-valued word vector indicating the absence/presence of\
    the corresponding word from the dictionary. The dictionary consists of 1433 unique\
    words.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "CORA",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Countries(CTUDataset):
    """
    Data of forest area for 247 countries.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Countries",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class CraftBeer(CTUDataset):
    """
    Craft beers labeled by styles and composition.\
    A separate dataset lists breweries by state.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "CraftBeer",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Credit(CTUDataset):
    """
    A bit more complex artificial database with loops.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Credit",
            cache_dir=cache_dir,
            time_col_dict={
                "charge": "charge_dt",
                "payment": "payment_dt",
            },
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        # TODO: Keep images and use them as features
        db.table_dict["member"].df.drop(columns=["photograph"], inplace=True)

        return db


class Dallas(CTUDataset):
    """
    Officer-involved shootings as disclosed by the Dallas Police Department.\
    Includes separate tables for officer and subject/suspect information.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Dallas",
            cache_dir=cache_dir,
            time_col_dict={"incidents": "date"},
            keep_original_keys=False,
        )


class DCG(CTUDataset):
    """
    The set of positive examples consists of all sentences of up to seven words that can\
    be generated by the DCG in Bratko's book (565 positive examples). The set of negative\
    examples was generated by randomly selecting one word in each positive example and\
    replacing it by a randomly selected word the leads to an incorrect sentence, according\
    to the grammar (565 negative examples).
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "DCG",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Diabetes(CTUDataset):
    """
    The Diabetes dataset consists of 19717 scientific publications from PubMed database\
    pertaining to diabetes classified into one of three classes. The citation network\
    consists of 44338 links. Each publication in the dataset is described by a TF/IDF\
    weighted word vector from a dictionary which consists of 500 unique words.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "PubMed_Diabetes",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Dunur(CTUDataset):
    """
    Dunur is a relatedness of two people due to marriage such that A is dunur of B if\
    a child of A is married to a child of B.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Dunur",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Elti(CTUDataset):
    """
    Elti is a relatedness of two people due to marriage such that A is elti of B if\
    A's husband is a brother of B's husband.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Elti",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Employee(CTUDataset):
    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "employee",
            cache_dir=cache_dir,
            time_col_dict={
                "dept_emp": "from_date",
                "dept_manager": "from_date",
                "employees": "hire_date",
                "salaries": "from_date",
                "titles": "from_date",
            },
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        # Drop redundant key columns
        db.table_dict["titles"].df.drop(columns=["emp_no"], inplace=True)
        db.table_dict["dept_manager"].df.drop(columns=["emp_no", "dept_no"], inplace=True)
        db.table_dict["departments"].df.drop(columns=["dept_no"], inplace=True)
        db.table_dict["salaries"].df.drop(columns=["emp_no"], inplace=True)
        db.table_dict["dept_emp"].df.drop(columns=["emp_no", "dept_no"], inplace=True)
        db.table_dict["employees"].df.drop(columns=["emp_no"], inplace=True)

        # Convert date columns to datetime and handle NaT values
        db.table_dict["titles"].df["to_date"] = pd.to_datetime(
            db.table_dict["titles"].df["to_date"], errors="coerce"
        )
        db.table_dict["dept_manager"].df["to_date"] = pd.to_datetime(
            db.table_dict["dept_manager"].df["to_date"], errors="coerce"
        )
        db.table_dict["salaries"].df["to_date"] = pd.to_datetime(
            db.table_dict["salaries"].df["to_date"], errors="coerce"
        )
        db.table_dict["dept_emp"].df["to_date"] = pd.to_datetime(
            db.table_dict["dept_emp"].df["to_date"], errors="coerce"
        )

        for table in db.table_dict.values():
            dtcols = table.df.select_dtypes(include=["datetime"])
            dtcols[dtcols > pd.Timestamp("2100-01-01")] = pd.NaT
            dtcols[dtcols < pd.Timestamp("1900-01-01")] = pd.NaT
            table.df[dtcols.columns] = dtcols

        return db


class ErgastF1(CTUDataset):
    """
    Ergast.com is a webservice that provides a database of Formula 1 races, \
    starting from the 1950 season until today. The dataset includes information \
    such as the time taken in each lap, the time taken for pit stops, the performance \
    in the qualifying rounds etc. of all Formula 1 races from 1950 to 2017.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "ErgastF1",
            cache_dir=cache_dir,
            time_col_dict={"races": "date"},
            keep_original_keys=False,
            keep_original_compound_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        # Convert time column to datetime
        db.table_dict["pitStops"].df["time"] = (
            db.table_dict["pitStops"].df["time"].fillna(pd.Timedelta(hours=12))
        )
        db.table_dict["pitStops"].df["time"] += db.table_dict["pitStops"].df.join(
            db.table_dict["races"].df["date"], on="FK_races_raceId", how="left"
        )["date"]
        db.table_dict["pitStops"].time_col = "time"

        # Merge date and time columns
        db.table_dict["races"].df["time"] = (
            db.table_dict["races"].df["time"].fillna(pd.Timedelta(hours=12))
        )
        db.table_dict["races"].df["date"] += db.table_dict["races"].df["time"]
        db.table_dict["races"].df.drop(columns=["time"], inplace=True)

        # Convert time columns to datetime
        db.table_dict["qualifying"].df["q1"] = pd.to_datetime(
            db.table_dict["qualifying"].df["q1"], format="%M:%S.%f", errors="coerce"
        )
        db.table_dict["qualifying"].df["q2"] = pd.to_datetime(
            db.table_dict["qualifying"].df["q2"], format="%M:%S.%f", errors="coerce"
        )
        db.table_dict["qualifying"].df["q3"] = pd.to_datetime(
            db.table_dict["qualifying"].df["q3"], format="%M:%S.%f", errors="coerce"
        )
        db.table_dict["results"].df["fastestLapTime"] = pd.to_datetime(
            db.table_dict["results"].df["fastestLapTime"],
            format="%M:%S.%f",
            errors="coerce",
        )

        db.table_dict["target"].df["date"] = db.table_dict["target"].df.join(
            db.table_dict["races"].df.set_index("__PK__")["date"], on="raceId", how="left"
        )["date"]
        db.table_dict["target"].time_col = "date"
        db.table_dict["target"].df.drop(columns=["raceId", "driverId"], inplace=True)

        return db


class Expenditures(CTUDataset):
    """
    The Consumer Expenditure Survey (CE) collects data on expenditures, income, \
    and demographics in the United States. The public-use microdata (PUMD) files \
    provide this information for individual respondents without any information that \
    could identify respondents. PUMD files include adjustments for information that is \
    missing because respondents were unwilling or unable to provide it. The files also \
    have been adjusted to reduce the likelihood of identifying respondents, either \
    directly or through inference. The task is to predict, whether the expenditure
    is a gift or not. Household ids change from year to year - this is a property of the \
    data source.
    
    Original source: www.bls.gov
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "ConsumerExpenditures",
            cache_dir=cache_dir,
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        # Remove remove split column
        db.table_dict["EXPENDITURES"].df.drop(columns=["IS_TRAINING"], inplace=True)

        return db


class Financial(CTUDataset):
    """
    PKDD'99 Financial dataset contains 606 successful and 76 not \
    successful loans along with their information and transactions. 
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "financial",
            cache_dir=cache_dir,
            time_col_dict={
                "account": "date",
                "card": "issued",
                "loan": "date",
                "trans": "date",
            },
            keep_original_keys=False,
        )


class FNHK(CTUDataset):
    """
    Anonymised data from a hospital in Hradec Kralove, Czech Republic, \
    about treatment and medication.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "FNHK",
            cache_dir=cache_dir,
            time_col_dict={
                "pripady": "Datum_prijeti",
                "vykony": "Datum_provedeni_vykonu",
                "zup": "Datum_provedeni_vykonu",
            },
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        # Drop redundant key columns
        db.table_dict["pripady"].df.drop(columns=["Identifikace_pripadu"], inplace=True)
        db.table_dict["vykony"].df.drop(columns=["Identifikace_pripadu"], inplace=True)
        db.table_dict["zup"].df.drop(columns=["Identifikace_pripadu"], inplace=True)

        return db


class FTP(CTUDataset):
    """
    PAKDD'15 Data Mining Competition: The data were obtained from simulations of product\
    viewing activities of users with known gender. The data closely follow the real-life\
    distribution in that regard.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "ftp",
            cache_dir=cache_dir,
            time_col_dict={"session": "start_time"},
            keep_original_keys=False,
        )


class Geneea(CTUDataset):
    """
    Data on deputies and senators in the Czech Republic.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "geneea",
            cache_dir=cache_dir,
            time_col_dict={"hl_hlasovani": "datum", "omluvy": "den"},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        # Combine date and time columns
        db.table_dict["hl_hlasovani"].df["cas"] = (
            db.table_dict["hl_hlasovani"].df["datum"]
            + db.table_dict["hl_hlasovani"].df["cas"]
        )

        db.table_dict["omluvy"].df["den"] = pd.to_datetime(
            db.table_dict["omluvy"].df["den"]
        )

        return db


class Genes(CTUDataset):
    """
    KDD Cup 2001 prediction of gene/protein function and localization.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "genes",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class GOSales(CTUDataset):
    """
    GO Sales dataset from IBM contains information about daily sales, methods, retailers,\
    and products of a fictitious outdoor equipment retail chain “Great Outdoors” (GO).
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "GOSales",
            cache_dir=cache_dir,
            time_col_dict={"go_daily_sales": "Date"},
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        fk_col, fk_name = self._reindex_fk(
            {name: t.df for name, t in db.table_dict.items()},
            "go_daily_sales",
            ["Product number"],
            "go_products",
            ["Product number"],
        )
        db.table_dict["go_daily_sales"].df[fk_name] = fk_col
        db.table_dict["go_daily_sales"].fkey_col_to_pkey_table[fk_name] = "go_products"

        fk_col, fk_name = self._reindex_fk(
            {name: t.df for name, t in db.table_dict.items()},
            "go_daily_sales",
            ["Retailer code"],
            "go_retailers",
            ["Retailer code"],
        )
        db.table_dict["go_daily_sales"].df[fk_name] = fk_col
        db.table_dict["go_daily_sales"].fkey_col_to_pkey_table[fk_name] = "go_retailers"

        db.table_dict["go_products"].df.drop(columns=["Product number"], inplace=True)
        db.table_dict["go_retailers"].df.drop(columns=["Retailer code"], inplace=True)
        db.table_dict["go_daily_sales"].df.drop(
            columns=["Retailer code", "Product number"], inplace=True
        )

        db.table_dict.pop("go_1k")

        return db


class Grants(CTUDataset):
    """
    This dataset includes funding grants from the National Science Foundation.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Grants",
            cache_dir=cache_dir,
            time_col_dict={
                "awards": "award_effective_date",
                "investigator_awards": "start_date",
            },
            keep_original_keys=False,
        )


class Hepatitis(CTUDataset):
    """
    PKDD'02 Hepatitis dataset describes 206 instances of Hepatitis B \
    (contrasting them against 484 cases of Hepatitis C).
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Hepatitis_std",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Hockey(CTUDataset):
    """
    The Hockey Database, in addition to the NHL, covers the following early and\
    alternative leagues: NHA, PCHA, WCHL and WHA. It contains individual and team\
    statistics from 1909-10 through the 2011-12 season.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Hockey",
            cache_dir=cache_dir,
            time_col_dict={
                "CombinedShutouts": "date",
            },
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict["CombinedShutouts"].df = db.table_dict["CombinedShutouts"].df.rename(
            columns={"date": "day"}
        )
        db.table_dict["CombinedShutouts"].df["date"] = pd.to_datetime(
            db.table_dict["CombinedShutouts"].df[["year", "month", "day"]]
        )
        db.table_dict["CombinedShutouts"].df.drop(
            columns=["year", "month", "day"], inplace=True
        )

        for table in db.table_dict.values():
            if "year" in table.df.columns:
                table.df["year"] = pd.to_datetime(table.df["year"], format="%Y")
                table.time_col = "year"

        # Remove tables with no foreign keys
        db.table_dict.pop("abbrev")
        db.table_dict.pop("AwardsMisc")
        # TODO: Possibly keep the table and fix FK to Master
        db.table_dict.pop("HOF")

        return db


class IMDb(CTUDataset):
    """
    The IMDb database: moderately large, real database of movies.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "imdb_ijs",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict["movies"].df["year"] = pd.to_datetime(
            db.table_dict["movies"].df["year"], format="%Y"
        )
        db.table_dict["movies"].time_col = "year"

        return db


class Lahman(CTUDataset):
    """
    Lahman's baseball database contains complete batting and pitching statistics\
    from 1871 to 2014, plus fielding statistics, standings, team stats, managerial\
    records, post-season data, and more.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "lahman_2014",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        for table in db.table_dict.values():
            if "yearID" in table.df.columns:
                table.df["yearID"] = pd.to_datetime(
                    table.df["yearID"], format="%Y", errors="coerce"
                )
                table.time_col = "yearID"

        return db


class LegalActs(CTUDataset):
    """
    Bulgarian court decision metadata.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "legalActs",
            cache_dir=cache_dir,
            time_col_dict={
                "legalacts": "StartDate",
            },
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        # Remove scrape fix table
        db.table_dict.pop("scrapefix")

        return db


class Mesh(CTUDataset):
    """
    This domain is about finite element methods in engineering where the number of\
    elements in the mesh model can vary between 1 and 17.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Mesh",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )


class Mondial(CTUDataset):
    """
    A geography dataset from University of Göttingen describes 114 Christian countries\
    and 71 non-Christian countries.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Mondial",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict["organization"].df["Established"] = (
            db.table_dict["organization"].df["Established"].dt.year
        )
        db.table_dict["politics"].df["Independence"] = (
            db.table_dict["politics"].df["Independence"].dt.year
        )

        return db


class Mooney(CTUDataset):
    """
    The dataset describes a family composed of 86 people across 5 generations.\
    The family dataset includes 744 positive instances and 1488 randomly generated\
    negative instances.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Mooney_Family",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )


class MovieLens(CTUDataset):
    """
    MovieLens data set from the UC Irvine machine learning repository.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "imdb_MovieLens",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class MuskSmall(CTUDataset):
    """
    The Musk database describes molecules occurring in different conformations.\
    Each molecule is either musk or non-musk and one of the conformations determines\
    this property. Such a problem is known as a multiple-instance problem, and is modeled\
    by two tables molecule and conformation, joined by a one-to-many association.\
    Confirmation contains a molecule identifier plus 166 continuous features.\
    Molecule just contains the identifier and the class. There are two versions of\
    the dataset, MuskSmall, containing 92 molecules and 476 confirmations, and MuskLarge,\
    containing 102 molecules and 6598 confirmations.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "MuskSmall",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class MuskLarge(CTUDataset):
    """
    The Musk database describes molecules occurring in different conformations.\
    Each molecule is either musk or non-musk and one of the conformations determines\
    this property. Such a problem is known as a multiple-instance problem, and is modeled\
    by two tables molecule and conformation, joined by a one-to-many association.\
    Confirmation contains a molecule identifier plus 166 continuous features.\
    Molecule just contains the identifier and the class. There are two versions of\
    the dataset, MuskSmall, containing 92 molecules and 476 confirmations, and MuskLarge,\
    containing 102 molecules and 6598 confirmations.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "MuskLarge",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Mutagenesis(CTUDataset):
    """
    The dataset comprises of 230 molecules trialed for mutagenicity on\
    Salmonella typhimurium. A subset of 188 molecules is learnable using linear regression.\
    This subset was later termed the ”regression friendly” dataset. The remaining subset\
    of 42 molecules is named the ”regression unfriendly” dataset.\
    Note that authors use this dataset with a variable set of the background knowledge\
    and consequently, the reported accuracies do not have to be directly comparable.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "mutagenesis",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class NBA(CTUDataset):
    """
    A database with information about basketball matches from the National Basketball\
    Association. Lists Players, Teams, and matches with action counts for each player.
    
    TODO: contains value errors
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "NBA",
            cache_dir=cache_dir,
            time_col_dict={"Game": "Date"},
            keep_original_keys=False,
        )


class NCAA(CTUDataset):
    """
    NCAA Basketball Tournament.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "NCAA",
            cache_dir=cache_dir,
            time_col_dict={"seasons": "dayzero"},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        for table in db.table_dict.values():
            for fk, ref in table.fkey_col_to_pkey_table.items():
                if ref != "seasons":
                    continue
                table.df["date"] = pd.merge(
                    left=db.table_dict["seasons"].df,
                    right=table.df,
                    left_on="__PK__",
                    right_on=fk,
                    how="right",
                )["dayzero"]

                if "daynum" in table.df.columns:
                    table.df["date"] += pd.to_timedelta(table.df["daynum"], unit="D")
                    table.df.drop(columns=["daynum"], inplace=True)

                table.time_col = "date"
                break

        db.table_dict["target"].df.drop(columns=["pred", "team_id2_wins"], inplace=True)

        return db


class Northwind(CTUDataset):
    """
    The Northwind database contains the sales data for a fictitious company called\
    Northwind Traders, which imports and exports specialty foods from around the world.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "northwind",
            cache_dir=cache_dir,
            time_col_dict={"Orders": "OrderDate", "Employees": "HireDate"},
            keep_original_keys=False,
        )


class Pima(CTUDataset):
    """
    The National Institute of Diabetes and Digestive and Kidney Diseases conducted\
    a study on 768 adult female Pima Indians living near Phoenix.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Pima",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class PremiereLeague(CTUDataset):
    """
    A database with information about football matches from the UK Premier League.\
    Lists Players, Teams, and matches with action counts for each player.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "PremierLeague",
            cache_dir=cache_dir,
            time_col_dict={"Matches": "Date"},
            keep_original_keys=False,
        )


class Restbase(CTUDataset):
    """
    A database of restaurants in the San Francisco Bay Area.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "restbase",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict["location"].df.drop(columns=["id_restaurant"], inplace=True)
        db.table_dict["generalinfo"].df.drop(columns=["id_restaurant"], inplace=True)

        return db


class Sakila(CTUDataset):
    """
    The Sakila sample database is designed to represent a DVD rental store.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "sakila",
            cache_dir=cache_dir,
            time_col_dict={"payment": "payment_date", "rental": "rental_date"},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        fk_name = "FK_film_film_id"
        db.table_dict["film_text"].df[fk_name] = db.table_dict["film_text"].df["__PK__"]
        db.table_dict["film_text"].fkey_col_to_pkey_table[fk_name] = "film"

        return db


class Sales(CTUDataset):
    """
    A simple artificial database in star schema.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "SalesDB",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class SameGen(CTUDataset):
    """
    Small database of family relations.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Same_gen",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )


class SAP(CTUDataset):
    """
    Syntetic dataset containing information about sales of a Credit++.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "SAP",
            cache_dir=cache_dir,
            time_col_dict={"Sales": "EVENT_DATE"},
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        mailings_1_2 = db.table_dict["Mailings1_2"].df
        mailings_1_2.drop(columns=["KxIndex", "REFID"], inplace=True)
        fk_name = [
            k
            for k, v in db.table_dict["Mailings1_2"].fkey_col_to_pkey_table.items()
            if v == "Customers"
        ][0]

        mailings_3 = db.table_dict["mailings3"].df
        fk_col, _ = self._reindex_fk(
            {tn: t.df for tn, t in db.table_dict.items()},
            "mailings3",
            ["REFID"],
            "Customers",
            ["ID"],
        )
        mailings_3[fk_name] = fk_col
        mailings_3.drop(columns=["REFID"], inplace=True)
        mailings_3["__PK__"] += mailings_1_2["__PK__"].max() + 1

        db.table_dict["Customers"].df.drop(columns=["ID", "GEOID"], inplace=True)
        db.table_dict["Sales"].df.drop(columns=["EVENTID", "REFID"], inplace=True)
        db.table_dict["Demog"].df.drop(columns=["GEOID"], inplace=True)

        db.table_dict.pop("Mailings1_2")
        db.table_dict.pop("mailings3")

        mailings = pd.concat([mailings_1_2, mailings_3], axis=0)
        db.table_dict["Mailings"] = Table(
            df=mailings,
            fkey_col_to_pkey_table={fk_name: "Customers"},
            pkey_col="__PK__",
            time_col="REF_DATE",
        )

        return db


class Satellite(CTUDataset):
    """
    Communications satellite data.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "SAT",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        for table in db.table_dict.values():
            rm_fks = []
            for fk, ref in table.fkey_col_to_pkey_table.items():
                if ref in ["trfl", "class"]:
                    table.df.drop(columns=[fk], inplace=True)
                    rm_fks.append(fk)
            for fk in rm_fks:
                table.fkey_col_to_pkey_table.pop(fk)

        db.table_dict.pop("trfl")
        db.table_dict.pop("class")

        return db


class Seznam(CTUDataset):
    """
    Seznam.cz is a web portal and search engine in the Czech Republic.\
    The data represent online advertisement expenditures from Seznam's "wallet".
    
    Table description: 
        client: location and domain field of the client (anonymized)
        dobito: prepaid into a wallet in Czech currency 
        probehnuto: charged from the wallet in Czech currency
        probehnuto_mimo_penezenku: charged in Czech currency, but not from the wallet.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Seznam",
            cache_dir=cache_dir,
            time_col_dict={
                "dobito": "month_year_datum_transakce",
                "probehnuto": "month_year_datum_transakce",
                "probehnuto_mimo_penezenku": "Month_Year",
            },
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict["probehnuto_mimo_penezenku"].df.rename(
            columns={"Month/Year": "Month_Year"}, inplace=True
        )

        return db


class SFScores(CTUDataset):
    """
    The San Francisco Dept. of Public Health's database of eateries, inspections\
    of those eateries, and violations found during the inspections. The scores of inspections\
    range from 1 to 100, where 100 means that the establishment meets all required standards.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "SFScores",
            cache_dir=cache_dir,
            time_col_dict={
                "inspections": "date",
                "violations": "date",
            },
            keep_original_keys=False,
        )


class Shakespeare(CTUDataset):
    """
    The Open Source Shakespeare is a collection of Shakespeare's complete works.\
    This is a much more interesting data set than some boring imaginary online retailer.\
    In this dataset, people die!
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Shakespeare",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class Stats(CTUDataset):
    """
    An anonymized dump of all user-contributed content on the Stats Stack Exchange network.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "stats",
            cache_dir=cache_dir,
            time_col_dict={
                "badges": "Date",
                "comments": "CreationDate",
                "postHistory": "CreationDate",
                "postLinks": "CreationDate",
                "posts": "CreaionDate",
                "users": "CreationDate",
                "votes": "CreationDate",
            },
            keep_original_keys=False,
        )


class StudentLoan(CTUDataset):
    """
    Student Loan contains data about students enrollment and employment status, and\
    the aim is to find rules that define a students' obligation for paying his/her loan back.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Student_loan",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )


class Thrombosis(CTUDataset):
    """
    PKDD'99 Medical dataset describes 41 patients with Thrombosis.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "medical",
            cache_dir=cache_dir,
            time_col_dict={"Laboratory": "Date"},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        db.table_dict["Examination"].df = db.table_dict["Examination"].df[
            db.table_dict["Examination"].df["Examination Date"].notna()
        ]

        return db


class Toxicology(CTUDataset):
    """
    Predictive Toxicology Challenge (2000) consists of more than three hundreds of organic\
    molecules marked according to their carcinogenicity on male and female mice and rats.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Toxicology",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class TPCC(CTUDataset):
    """
    TPC-C is the benchmark published by the Transaction Processing Performance Council\
    (TPC) for Online Transaction Processing (OLTP).
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpcc",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=True,
        )


class TPCD(CTUDataset):
    """
    TPC-D represents a broad range of decision support (DS) applications that require\
    complex, long running queries against large complex data structures.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpcd",
            cache_dir=cache_dir,
            time_col_dict={
                "dss_lineitem": "l_shipdate",
                "dss_order": "o_orderdate",
            },
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        # Add missing foreign keys
        fk_col, fk_name = self._reindex_fk(
            {tn: t.df for tn, t in db.table_dict.items()},
            "dss_partsupp",
            ["ps_partkey"],
            "dss_part",
            ["p_partkey"],
        )
        db.table_dict["dss_partsupp"].fkey_col_to_pkey_table[fk_name] = "dss_part"
        db.table_dict["dss_partsupp"].df[fk_name] = fk_col

        # Drop redundant key columns
        db.table_dict["dss_lineitem"].df.drop(columns=["l_orderkey"], inplace=True)
        db.table_dict["dss_part"].df.drop(columns=["p_partkey"], inplace=True)
        db.table_dict["dss_partsupp"].df.drop(
            columns=["ps_partkey", "ps_suppkey"], inplace=True
        )
        db.table_dict["dss_region"].df.drop(columns=["r_regionkey"], inplace=True)
        db.table_dict["dss_nation"].df.drop(
            columns=["n_nationkey", "n_regionkey"], inplace=True
        )
        db.table_dict["dss_customer"].df.drop(
            columns=["c_custkey", "c_nationkey"], inplace=True
        )
        db.table_dict["dss_supplier"].df.drop(
            columns=["s_suppkey", "s_nationkey"], inplace=True
        )
        db.table_dict["dss_order"].df.drop(
            columns=["o_orderkey", "o_custkey"], inplace=True
        )

        return db


class TPCDS(CTUDataset):
    """
    TPC-DS is the new decision support benchmark that models several generally applicable\
    aspects of a decision support system, including queries and data maintenance. Although\
    the underlying business model of TPC-DS is a retail product supplier, the database\
    schema, data population, queries, data maintenance model and implementation rules have\
    been designed to be broadly representative of modern decision support systems.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpcds",
            cache_dir=cache_dir,
            time_col_dict={
                "store": "s_rec_start_date",
            },
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        # Replace date and time tables with simple datetime columns
        time_df = db.table_dict["time_dim"].df
        date_df = db.table_dict["date_dim"].df

        def to_time(x):
            return pd.Timedelta(
                hours=x["t_hour"], minutes=x["t_minute"], seconds=x["t_second"]
            )

        time_df["time_delta"] = db.table_dict["time_dim"].df.apply(to_time, axis=1)

        date_fk_dict = {
            tn: [
                fk for fk, fk_tn in t.fkey_col_to_pkey_table.items() if fk_tn == "date_dim"
            ]
            for tn, t in db.table_dict.items()
        }

        time_col_dict = {
            "store": "s_rec_start_date",
            "customer": "c_first_sales_date_sk",
            "web_page": "wp_creation_date_sk",
            "inventory": "inv_date_sk",
            "catalog_page": "cp_start_date_sk",
            "promotion": "p_start_date_sk",
            "web_site": "web_open_date_sk",
            "item": "i_rec_start_date",
            "store_returns": "sr_returned_date_sk",
            "catalog_sales": "cs_sold_date_sk",
            "call_center": "cc_open_date_sk",
            "web_sales": "ws_ship_date_sk",
        }

        time_fk_dict = {
            "store_returns": ("FK_time_dim_sr_return_time_sk", "sr_returned_date_sk"),
            "catalog_returns": ("FK_time_dim_cr_returned_time_sk", "cr_returned_date_sk"),
            "web_returns": ("FK_time_dim_wr_returned_time_sk", "wr_returned_date_sk"),
            "store_sales": ("FK_time_dim_ss_sold_time_sk", "ss_sold_date_sk"),
            "catalog_sales": ("FK_time_dim_cs_sold_time_sk", "cs_sold_date_sk"),
            "web_sales": ("FK_time_dim_ws_sold_time_sk", "ws_sold_date_sk"),
        }

        for t_name, fks in date_fk_dict.items():
            for fk in fks:
                db.table_dict[t_name].df[fk.removeprefix("FK_date_dim_")] = db.table_dict[
                    t_name
                ].df.join(date_df["d_date"], on=fk, how="left")["d_date"]
                db.table_dict[t_name].df.drop(columns=[fk], inplace=True)
                db.table_dict[t_name].fkey_col_to_pkey_table.pop(fk)
            time_col = time_col_dict.get(t_name, None)
            if time_col is not None:
                db.table_dict[t_name].time_col = time_col

        for t_name, (fk, date_col) in time_fk_dict.items():
            db.table_dict[t_name].df[date_col] += (
                db.table_dict[t_name]
                .df[["__PK__", fk]]
                .fillna(0)
                .join(time_df["time_delta"], on=fk, how="left")["time_delta"]
            )
            db.table_dict[t_name].df.drop(columns=[fk], inplace=True)
            db.table_dict[t_name].fkey_col_to_pkey_table.pop(fk)

        db.table_dict.pop("time_dim")
        db.table_dict.pop("date_dim")

        return db


class TPCH(CTUDataset):
    """
    TPC-H is the benchmark published by the Transaction Processing Performance Council\
    (TPC) for decision support.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpch",
            cache_dir=cache_dir,
            time_col_dict={"lineitem": "l_shipdate", "orders": "o_orderdate"},
            keep_original_keys=False,
        )


class Triazine(CTUDataset):
    """
    A pyrimidine QSAR dataset. The the goal is to predict the inhibition\
    of dihydrofolate reductase by pyrimidines.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Triazine",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class UWCSE(CTUDataset):
    """
    This dataset lists facts about the Department of Computer Science and Engineering\
    at the University of Washington (UW-CSE), such as entities (e.g., Student, Professor)\
    and their relationships (i.e. AdvisedBy, Publication).
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "UW_std",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class VisualGenome(CTUDataset):
    """
    Visual Genome is a dataset, a knowledge base, an ongoing effort to connect\
    structured image concepts to language.
    
    TODO: add images
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "VisualGenome",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        for table in db.table_dict.values():
            rm_fks = []
            for fk, ref in table.fkey_col_to_pkey_table.items():
                if ref in ["ATT_CLASSES", "OBJ_CLASSES", "PRED_CLASSES"]:
                    col = ref.removesuffix("ES")

                    table.df[col] = pd.merge(
                        left=db.table_dict[ref].df,
                        right=table.df,
                        left_on="__PK__",
                        right_on=fk,
                        how="right",
                    )[col]
                    table.df.drop(columns=[fk], inplace=True)
                    rm_fks.append(fk)

            for fk in rm_fks:
                table.fkey_col_to_pkey_table.pop(fk)

        db.table_dict.pop("ATT_CLASSES")
        db.table_dict.pop("OBJ_CLASSES")
        db.table_dict.pop("PRED_CLASSES")

        return db


class VOC(CTUDataset):
    """
    VOC database provides a peephole view into the administrative system of an early\
    multi-national company, the Vereenigde geoctrooieerde Oostindische Compagnie (VOC for\
    short - The (Dutch) East Indian Company) established on March 20, 1602.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "voc",
            cache_dir=cache_dir,
            time_col_dict={"voyages": "departure_date"},
            keep_original_keys=False,
        )

    def customize_db(self, db: Database) -> Database:
        dtcols = db.table_dict["voyages"].df.select_dtypes(include="datetime")
        dtcols[dtcols < pd.Timestamp("1300-01-01")] = pd.NaT
        dtcols[dtcols > pd.Timestamp("1900-01-01")] = pd.NaT
        min_date = dtcols.min().min()
        max_date = dtcols.max().max()
        delta: pd.Timedelta
        if min_date < TIMESTAMP_MIN:
            delta = TIMESTAMP_MIN - min_date
            delta = pd.Timedelta(days=delta.days + 1)
        elif max_date > TIMESTAMP_MAX:
            delta = TIMESTAMP_MAX - max_date
            delta = pd.Timedelta(days=delta.days - 1)
        dtcols += delta.to_numpy().astype(np.dtype("timedelta64[D]"))

        db.table_dict["voyages"].df[dtcols.columns] = dtcols

        return db


class Walmart(CTUDataset):
    """
    The sales of 111 potentially weather-sensitive products (like umbrellas, bread, and milk)\
    around the time of major weather events at 45 of their retail locations.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Walmart",
            cache_dir=cache_dir,
            time_col_dict={"weather": "date"},
            keep_original_keys=True,
        )

    def customize_db(self, db: Database) -> Database:
        # Convert sunrise and sunset to datetime
        weather_df = db.table_dict["weather"].df
        weather_df["sunrise"] = weather_df["date"] + weather_df["sunrise"].fillna(
            pd.Timedelta(hours=6)
        )
        weather_df["sunset"] = weather_df["date"] + weather_df["sunset"].fillna(
            pd.Timedelta(hours=18)
        )
        db.table_dict["weather"].df = weather_df

        table = db.table_dict.pop("train")
        table.time_col = "date"
        db.table_dict["train_table"] = table

        return db


class WebKP(CTUDataset):
    """
    The WebKB dataset consists of 877 scientific publications classified into one of five\
    classes. The citation network consists of 1608 links. Each publication in the dataset\
    is described by a 0/1-valued word vector indicating the absence/presence of the\
    corresponding word from the dictionary. The dictionary consists of 1703 unique words.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "WebKP",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )


class World(CTUDataset):
    """
    A database of 239 states and their cities.
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "world",
            cache_dir=cache_dir,
            time_col_dict={},
            keep_original_keys=False,
        )



from typing import Callable, Dict, List, Optional
from numpy.typing import NDArray

import numpy as np

from relbench.base import Table

from .ctu_entity_task_base import CTUBaseEntityTask

__all__ = ["CTUEntityTask"]


class CTUEntityTask(CTUBaseEntityTask):
    def make_split(self, split: str, table: Table) -> Table:
        random_state = np.random.RandomState(seed=42)
        train_df = table.df.sample(frac=0.8, random_state=random_state)
        if split == "train":
            table.df = train_df
        else:
            table.df = table.df.drop(train_df.index)
            val_df = table.df.sample(frac=0.5, random_state=random_state)
            if split == "val":
                table.df = val_df
            else:
                table.df = table.df.drop(val_df.index)

        table.df = table.df[[self.entity_col, self.target_col]]
        table.time_col = None

        return table

    def evaluate(
        self,
        pred: NDArray,
        target_table: Optional[Table] = None,
        metrics: Optional[List[Callable[[NDArray, NDArray], float]]] = None,
    ) -> Dict[str, float]:
        if metrics is None:
            metrics = self.metrics

        if target_table is None:
            target_table = self.get_table("test", mask_input_cols=False)

        target = target_table.df[self.target_col].to_numpy()
        if len(pred) != len(target):
            raise ValueError(
                f"The length of pred and target must be the same (got "
                f"{len(pred)} and {len(target)}, respectively)."
            )

        return {fn.__name__: fn(target, pred) for fn in metrics}



from typing import Dict, Literal, Optional

import pandas as pd


from .db_dataset import DBDataset

__all__ = ["CTUDataset"]


# fmt: off
CTUDatabaseName = Literal[
    "Accidents", "AdventureWorks2014", "Airline", "Atherosclerosis", "AustralianFootball", 
    "Basketball_men", "Basketball_women", "Biodegradability", "Bupa", "Carcinogenesis", 
    "ccs", "CDESchools", "Chess", "CiteSeer", "classicmodels", "ConsumerExpenditures",
    "CORA", "Countries", "CraftBeer", "Credit", "cs", "Dallas", "DCG", "Dunur", "Elti",
    "employee", "ErgastF1", "Facebook", "financial", "FNHK", "ftp", "geneea", "genes",
    "GOSales", "Grants", "Hepatitis_std", "Hockey", "imdb_ijs", "KRK", "lahman_2014",
    "legalActs", "Mesh", "medical", "Mondial", "Mooney_Family", "imdb_MovieLens",
    "MuskLarge", "MuskSmall", "mutagenesis", "nations", "NBA", "NCAA", "northwind", "Pima",
    "PremierLeague", "PTE", "PubMed_Diabetes", "pubs", "Pyrimidine", "restbase",
    "sakila", "SalesDB", "Same_gen", "SAP", "SAT", "Seznam", "SFScores", "Shakespeare",
    "stats", "Student_loan", "Toxicology", "tpcc", "tpcd", "tpcds", "tpch", "trains",
    "Triazine", "university", "UTube", "UW_std", "VisualGenome", "voc", "Walmart", "WebKP",
    "world"
]
# fmt: on


class CTUDataset(DBDataset):
    # To be set by subclass if available.
    val_timestamp = pd.Timestamp.max.date()
    test_timestamp = pd.Timestamp.max.date()

    def __init__(
        self,
        database: CTUDatabaseName,
        cache_dir: Optional[str] = None,
        time_col_dict: Optional[Dict[str, str]] = None,
        keep_original_keys: bool = False,
        keep_original_compound_keys: bool = True,
    ):
        """Create a database dataset object.

        Args:
            database (CTUDatabaseName): The name of the database.
            cache_dir (str, optional): The directory to cache the dataset. Defaults to None.
            time_col_dict (Dict[str, str], optional): A dictionary mapping table names to time columns. Defaults to None.
            keep_original_keys (bool, optional): Whether to keep original primary and foreign keys \
                after duplication during re-indexing. This is useful when the keys contain information \
                beyond just their relationship to other rows. Defaults to False.
            keep_original_compound_keys (bool, optional): Whether to keep original compound primary \
                and foreign keys as they often contain useful data. Defaults to True.
        """
        self.database = database
        super().__init__(
            cache_dir=cache_dir,
            dialect="mariadb",
            driver="pymysql",
            user="guest",
            password="ctu-relational",
            host="relational.fel.cvut.cz",
            port=3306,
            database=database,
            time_col_dict=time_col_dict,
            keep_original_keys=keep_original_keys,
            keep_original_compound_keys=keep_original_compound_keys,
        )



from typing import Dict, Optional

import numpy as np
import pandas as pd

from torch_frame import stype

from relbench.base import Database, Table


TIMESTAMP_MIN = np.datetime64(pd.Timestamp.min.date())
TIMESTAMP_MAX = np.datetime64(pd.Timestamp.max.date())


def convert_timedelta(db: Database):
    """Converts timedelta columns to datetime columns."""

    for table in db.table_dict.values():
        timedeltas = table.df.select_dtypes(include=["timedelta"])
        if not timedeltas.empty:
            timedeltas = pd.Timestamp("1900-01-01") + timedeltas
            table.df[timedeltas.columns] = timedeltas


def standardize_table_dt(
    table: Table, col_to_stype: Optional[Dict[str, stype]] = None
) -> None:
    """Standartize datetime columns to UNIX timestamp (in datetime[ns] if possible)."""

    if col_to_stype is not None:
        for col, s in col_to_stype.items():
            if s == stype.timestamp:
                table.df[col] = table.df[col].astype(np.dtype("datetime64[ns]"))

    if table.time_col is not None:
        table.df[table.time_col] = table.df[table.time_col].astype(
            np.dtype("datetime64[ns]")
        )


def standardize_db_dt(db: Database, col_to_stype_dict: Dict[str, Dict[str, stype]] = {}):
    """Standartize datetime columns to UNIX timestamp (in datetime[ns] if possible)."""

    for tname, table in db.table_dict.items():
        standardize_table_dt(table, col_to_stype_dict.get(tname, {}))


__all__ = [
    "convert_timedelta",
    "standardize_db_dt",
    "standardize_table_dt",
    "TIMESTAMP_MIN",
    "TIMESTAMP_MAX",
]



from typing import Dict, Optional, overload

import re
import warnings

from sqlalchemy.dialects.mysql import types as mysql_types
from sqlalchemy import types as sql_types

import numpy as np

import pandas as pd
from pandas.api import types as pd_types
from dateutil.parser import ParserError

import inflect

from torch_frame import stype
from torch_frame.utils import infer_series_stype

from relbench.base import BaseTask, Database, Table, TaskType

from ctu_relational.datasets import DBDataset

ID_NAME_REGEX = re.compile(
    r"_id$|^id_|_id_|Id$|Id[^a-z]|[Ii]dentifier|IDENTIFIER|ID[^a-zA-Z]|ID$|[guGU]uid[^a-z]|[guGU]uid$|[GU]UID[^a-zA-Z]|[GU]UID$"
)

COMMON_NUMERIC_COLUMN_NAME_REGEX = re.compile(
    r"balance|amount|size|duration|frequency|count|cnt|votes|score|number|age|year|month|day",
    re.IGNORECASE,
)

FRACTION_DISTINCT_NONNULL_GUARANTEED_THRESHOLD = 0.05
"""
The fraction of distinct values to total count of non-null values,
which decides (in some situations) that type must be categorical.
If the fraction is below this threshold, marks the column as categorical.
"""

FRACTION_DISTINCT_NONNULL_IGNORE_THRESHOLD = 0.2
"""
The fraction of distinct values to total count of non-null values,
which decides (in some situations) that type cannot be categorical.
If the fraction exceeds this threshold, marks the column as something other than categorical.
"""

MAXIMUM_CARDINALITY_THRESHOLD = 1000
"""
Maximum number of cardinality that a categorical column can have.
This is present to limit incorrectly classified categorical columns.
"""

POSSIBLE_TIME_FORMATS = [
    "%Y-%m-%d %H:%M:%S",
    "%Y/%m/%d %H:%M:%S",
    "%Y-%m-%d %H:%M:%S.%f",
    "%Y-%m-%d",
    "%Y/%m/%d",
]
"""

"""

pd.to_datetime


def check_predetermined_types(
    ser: pd.Series, sql_type: Optional[sql_types.TypeEngine]
) -> Optional[stype]:
    if sql_type is None:
        sql_type = sql_types.NullType()
    sql_type_generic = sql_type.as_generic(allow_nulltype=True)

    if isinstance(sql_type, (mysql_types.LONGTEXT, mysql_types.MEDIUMTEXT)) or isinstance(
        sql_type_generic, (sql_types.Unicode)
    ):
        return stype.text_embedded

    if isinstance(sql_type_generic, (sql_types.Boolean)) or pd_types.is_bool_dtype(ser):
        return stype.categorical

    if isinstance(sql_type_generic, (sql_types.Numeric)) or pd_types.is_float_dtype(ser):
        return stype.numerical

    if isinstance(
        sql_type_generic, (sql_types.Date, sql_types.DateTime, sql_types.Time)
    ) or pd_types.is_datetime64_any_dtype(ser):
        return stype.timestamp

    return None


def guess_column_stype(
    ser: pd.Series, col_name: str = "", sql_type: Optional[sql_types.TypeEngine] = None
) -> stype:
    ser = ser.dropna()
    if ser.empty:
        return None

    if sql_type is None:
        sql_type = sql_types.NullType()
    sql_generic_type = sql_type.as_generic(allow_nulltype=True)

    # Handle list data with torch frame infer
    if isinstance(ser.iloc[0], (list, np.ndarray)):
        if isinstance(ser.iloc[0], np.ndarray):
            ser = ser.apply(np.ndarray.tolist)
        return infer_series_stype(ser)

    predeter_type = check_predetermined_types(ser, sql_type)
    if predeter_type is not None:
        return predeter_type

    # check if cardinality to total count ratio is below categorical threshold
    if _is_categorical(ser):
        return stype.categorical

    if isinstance(sql_generic_type, sql_types.Integer) or pd_types.is_numeric_dtype(ser):
        # check if there are too many distinct values compared to total
        if _is_not_categorical(ser):
            if _is_id_name(col_name):
                return None
            return stype.numerical

        # try matching based on common regex names
        if COMMON_NUMERIC_COLUMN_NAME_REGEX.search(col_name):
            return stype.numerical

        # check if the column name is plural - then it is probably a count
        if _is_plural(col_name):
            return stype.numerical

        return stype.categorical

    elif isinstance(
        sql_generic_type, (sql_types.String, sql_types.Text)
    ) or pd_types.is_string_dtype(ser):
        # if the column contains only empty strings, we can ignore it
        if (ser == "").all():
            return None

        # check if the values fit some of the time formats
        if _is_timestamp(ser):
            return stype.timestamp

        # check if there are too many distinct values compared to total
        if _is_not_categorical(ser):
            if _is_id_name(col_name):
                return None
            return stype.text_embedded

        return stype.categorical

    # no decision - omit
    return None


def guess_table_stypes(
    table: Table,
    table_schema: Optional[Dict[str, sql_types.TypeEngine]] = {},
    task: Optional[BaseTask] = None,
    ignore_none: bool = True,
) -> Dict[str, stype]:
    """
    Guess the stypes of columns in the table.

    Contains additional logic for foreign keys and filtering based on constructor input.
    """

    schema: Dict[str, stype] = {}

    for col in table.df.columns:
        if table.pkey_col is not None and col == table.pkey_col:
            schema[col] = None
            continue

        if table.time_col is not None and col == table.time_col:
            schema[col] = stype.timestamp
            continue

        if col in table.fkey_col_to_pkey_table:
            schema[col] = None
            continue

        if task is not None:
            if task.task_type == TaskType.BINARY_CLASSIFICATION:
                schema[col] = stype.categorical
            elif task.task_type == TaskType.MULTICLASS_CLASSIFICATION:
                schema[col] = stype.categorical
            elif task.task_type == TaskType.REGRESSION:
                schema[col] = stype.numerical
            elif task.task_type == TaskType.MULTILABEL_CLASSIFICATION:
                schema[col] = stype.multicategorical
            else:
                raise ValueError(f"Unknown task type {task.task_type.name}")
            continue

        guess = guess_column_stype(
            table.df[col], col_name=col, sql_type=table_schema.get(col, None)
        )
        if ignore_none and guess is None:
            continue

        schema[col] = guess
    return schema


@overload
def guess_schema(dataset: DBDataset) -> Dict[str, Dict[str, stype]]:
    schema = {}

    db = dataset.get_db(upto_test_timestamp=False)

    sql_schema = dataset.get_schema()

    for table_name, table in db.table_dict.items():
        schema[table_name] = guess_table_stypes(
            table, table_schema=sql_schema.get(table_name, {})
        )

    return schema


def guess_schema(
    db: Database, sql_schema: Optional[Dict[str, Dict[str, sql_types.TypeEngine]]] = {}
) -> Dict[str, Dict[str, stype]]:
    """Locate all database tables and all columns and run :py:method:`guess_column_type` for all of them.

    Returns the result as a :py:class:`Schema`.
    """
    schema = {}

    for table_name, table in db.table_dict.items():
        schema[table_name] = guess_table_stypes(
            table, table_schema=sql_schema.get(table_name, {})
        )

    return schema


def _is_not_categorical(ser: pd.Series) -> bool:
    cardinality = ser.unique().size
    n_nonnull = ser.count()

    return (
        cardinality / n_nonnull > FRACTION_DISTINCT_NONNULL_IGNORE_THRESHOLD
        or cardinality > MAXIMUM_CARDINALITY_THRESHOLD
    )


def _is_categorical(ser: pd.Series) -> bool:
    cardinality = ser.unique().size
    n_nonnull = ser.count()

    return (
        cardinality / n_nonnull <= FRACTION_DISTINCT_NONNULL_GUARANTEED_THRESHOLD
        and cardinality < MAXIMUM_CARDINALITY_THRESHOLD
    )


def _is_id_name(col_name: str) -> bool:
    return ID_NAME_REGEX.search(col_name) is not None


def _is_plural(s: str) -> bool:
    p = inflect.engine()
    return p.singular_noun(s) is not False


def _is_timestamp(ser: pd.Series) -> bool:
    r"""Check if a series is a timestamp.

    Taken from torch_frame
    https://github.com/pyg-team/pytorch-frame/blob/1456fab68ad291e8fe0572716704c69f36409bf6/torch_frame/utils/infer_stype.py#L21
    """
    for time_format in POSSIBLE_TIME_FORMATS:
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                pd.to_datetime(ser, format=time_format)
            return True
        except (ValueError, ParserError, TypeError):
            pass
    return False


__all__ = [
    "check_predetermined_types",
    "guess_column_stype",
    "guess_table_stypes",
    "guess_schema",
]



import pandas as pd
from relbench.base import TaskType

from .ctu_entity_task_classic import CTUEntityTask
from .ctu_entity_task_temporal import CTUEntityTaskTemporal
from .ctu_link_task import CTULinkTask


class AccidentsOriginalTask(CTUEntityTask):
    entity_table = "nesreca"
    target_col = "klas_nesreca"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class AccidentsTemporalTask(CTUEntityTaskTemporal):
    entity_table = "nesreca"
    target_col = "klas_nesreca"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("2004-03-01")
    test_timestamp = pd.Timestamp("2005-03-01")


class AdventureWorksOriginalTask(CTUEntityTask):
    entity_table = "SalesOrderHeader"
    target_col = "TotalDue"
    task_type = TaskType.REGRESSION


class AdventureWorksTemporalTask(CTUEntityTaskTemporal):
    entity_table = "SalesOrderHeader"
    target_col = "TotalDue"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2014-03-01")
    test_timestamp = pd.Timestamp("2014-05-01")


# TODO: remove duplicate target cols
class AirlineOriginalTask(CTUEntityTask):
    entity_table = "On_Time_On_Time_Performance_2016_1"
    target_col = "ArrDelay"
    task_type = TaskType.REGRESSION


# TODO: remove duplicate target cols
class AirlineTemporalTask(CTUEntityTaskTemporal):
    entity_table = "On_Time_On_Time_Performance_2016_1"
    target_col = "ArrDelay"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2016-01-18")
    test_timestamp = pd.Timestamp("2016-01-25")


class AtherosclerosisOriginalTask(CTUEntityTask):
    entity_table = "Entry"
    target_col = "STAV"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class BasketballMenOriginalTask(CTUEntityTask):
    entity_table = "teams"
    target_col = "rank"
    task_type = TaskType.REGRESSION


class BasketballWomenOriginalTask(CTUEntityTask):
    entity_table = "teams"
    target_col = "playoff"
    task_type = TaskType.BINARY_CLASSIFICATION


class BiodegradabilityOriginalTask(CTUEntityTask):
    entity_table = "molecule"
    target_col = "activity"
    task_type = TaskType.REGRESSION


class BupaOriginalTask(CTUEntityTask):
    entity_table = "bupa"
    target_col = "arg2"
    task_type = TaskType.BINARY_CLASSIFICATION


class CarcinogenesisOriginalTask(CTUEntityTask):
    entity_table = "canc"
    target_col = "class"
    task_type = TaskType.BINARY_CLASSIFICATION


class CDESchoolsOriginalTask(CTUEntityTask):
    entity_table = "satscores"
    target_col = "PctGE1500"
    task_type = TaskType.REGRESSION


class ChessOriginalTask(CTUEntityTask):
    entity_table = "game"
    target_col = "game_result"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class ClassicModelsOriginalTask(CTUEntityTask):
    entity_table = "payments"
    target_col = "amount"
    task_type = TaskType.REGRESSION


class ClassicModelsTemporalTask(CTUEntityTaskTemporal):
    entity_table = "payments"
    target_col = "amount"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2004-11-01")
    test_timestamp = pd.Timestamp("2005-02-01")


class CORAOriginalTask(CTUEntityTask):
    entity_table = "paper"
    target_col = "class_label"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class CountriesOriginalTask(CTUEntityTask):
    entity_table = "target"
    target_col = "2012"
    task_type = TaskType.REGRESSION


class CraftBeerOriginalTask(CTUEntityTask):
    entity_table = "breweries"
    target_col = "state"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class CreditOriginalTask(CTULinkTask):
    entity_table = "member"
    link_table = "region"
    task_type = TaskType.LINK_PREDICTION
    val_timestamp = pd.Timestamp("1999-09-01")
    test_timestamp = pd.Timestamp("1999-10-01")


class DallasOriginalTask(CTUEntityTask):
    entity_table = "incidents"
    target_col = "subject_statuses"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class DallasTemporalTask(CTUEntityTaskTemporal):
    entity_table = "incidents"
    target_col = "subject_statuses"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("2014-01-01")
    test_timestamp = pd.Timestamp("2015-01-01")


class DCGOriginalTask(CTUEntityTask):
    entity_table = "sentences"
    target_col = "class"
    task_type = TaskType.BINARY_CLASSIFICATION


class DiabetesOriginalTask(CTUEntityTask):
    entity_table = "paper"
    target_col = "class_label"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class DunurOriginalTask(CTULinkTask):
    entity_table = "target"
    target_col = "is_dunur"
    task_type = TaskType.LINK_PREDICTION


class EltiOriginalTask(CTULinkTask):
    entity_table = "target"
    target_col = "is_elti"
    task_type = TaskType.LINK_PREDICTION


class EmployeeOriginalTask(CTUEntityTask):
    entity_table = "salaries"
    target_col = "salary"
    task_type = TaskType.REGRESSION


class EmployeeTemporalTask(CTUEntityTaskTemporal):
    entity_table = "salaries"
    target_col = "salary"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2000-01-01")
    test_timestamp = pd.Timestamp("2001-01-01")


class ErgastF1OriginalTask(CTUEntityTask):
    entity_table = "target"
    target_col = "win"
    task_type = TaskType.BINARY_CLASSIFICATION


class ErgastF1TemporalTask(CTUEntityTaskTemporal):
    entity_table = "target"
    target_col = "win"
    task_type = TaskType.BINARY_CLASSIFICATION
    val_timestamp = pd.Timestamp("2014-01-01")
    test_timestamp = pd.Timestamp("2016-01-01")


class ExpendituresOriginalTask(CTUEntityTask):
    entity_table = "EXPENDITURES"
    target_col = "GIFT"
    task_type = TaskType.BINARY_CLASSIFICATION


class FinancialOriginalTask(CTUEntityTask):
    entity_table = "loan"
    target_col = "status"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class FinancialTemporalTask(CTUEntityTaskTemporal):
    entity_table = "loan"
    target_col = "status"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("1998-01-01")
    test_timestamp = pd.Timestamp("1998-07-01")


class FNHKOriginalTask(CTUEntityTask):
    entity_table = "pripady"
    target_col = "Delka_hospitalizace"
    task_type = TaskType.REGRESSION


class FNHKTemporalTask(CTUEntityTaskTemporal):
    entity_table = "pripady"
    target_col = "Delka_hospitalizace"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2014-09-01")
    test_timestamp = pd.Timestamp("2014-11-01")


class FTPOriginalTask(CTUEntityTask):
    entity_table = "session"
    target_col = "gender"
    task_type = TaskType.BINARY_CLASSIFICATION


class FTPTemporalTask(CTUEntityTaskTemporal):
    entity_table = "session"
    target_col = "gender"
    task_type = TaskType.BINARY_CLASSIFICATION
    val_timestamp = pd.Timestamp("2014-12-15")
    test_timestamp = pd.Timestamp("2014-12-19")


class GeneeaOriginalTask(CTUEntityTask):
    entity_table = "hl_hlasovani"
    target_col = "vysledek"
    task_type = TaskType.BINARY_CLASSIFICATION


class GeneeaTemporalTask(CTUEntityTaskTemporal):
    entity_table = "hl_hlasovani"
    target_col = "vysledek"
    task_type = TaskType.BINARY_CLASSIFICATION
    val_timestamp = pd.Timestamp("2015-06-01")
    test_timestamp = pd.Timestamp("2015-10-01")


class GenesOriginalTask(CTUEntityTask):
    entity_table = "Classification"
    target_col = "Localization"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class GOSalesOriginalTask(CTUEntityTask):
    entity_table = "go_daily_sales"
    target_col = "Quantity"
    task_type = TaskType.REGRESSION


class GOSalesTemporalTask(CTUEntityTaskTemporal):
    entity_table = "go_daily_sales"
    target_col = "Quantity"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2017-10-01")
    test_timestamp = pd.Timestamp("2018-03-01")


class GrantsOriginalTask(CTUEntityTask):
    entity_table = "awards"
    target_col = "award_amount"
    task_type = TaskType.REGRESSION


class GrantsTemporalTask(CTUEntityTaskTemporal):
    entity_table = "awards"
    target_col = "award_amount"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2010-10-01")
    test_timestamp = pd.Timestamp("2014-01-01")


class HepatitisOriginalTask(CTUEntityTask):
    entity_table = "dispat"
    target_col = "Type"
    task_type = TaskType.BINARY_CLASSIFICATION


class HockeyOriginalTask(CTUEntityTask):
    entity_table = "Master"
    target_col = "shootCatch"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class HockeyTemporalTask(CTUEntityTaskTemporal):
    entity_table = "Master"
    target_col = "shootCatch"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("2007-01-01")
    test_timestamp = pd.Timestamp("2009-01-01")


class IMDbOriginalTask(CTUEntityTask):
    entity_table = "actors"
    target_col = "gender"
    task_type = TaskType.BINARY_CLASSIFICATION


class IMDbTemporalTask(CTUEntityTaskTemporal):
    entity_table = "actors"
    target_col = "gender"
    task_type = TaskType.BINARY_CLASSIFICATION
    val_timestamp = pd.Timestamp("1998-01-01")
    test_timestamp = pd.Timestamp("2002-01-01")


class LahmanOriginalTask(CTUEntityTask):
    entity_table = "salaries"
    target_col = "salary"
    task_type = TaskType.REGRESSION


class LahmanTemporalTask(CTUEntityTaskTemporal):
    entity_table = "salaries"
    target_col = "salary"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2010-01-01")
    test_timestamp = pd.Timestamp("2012-01-01")


class LegalActsOriginalTask(CTUEntityTask):
    entity_table = "legalacts"
    target_col = "ActKind"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class LegalActsTemporalTask(CTUEntityTaskTemporal):
    entity_table = "legalacts"
    target_col = "ActKind"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("2012-02-01")
    test_timestamp = pd.Timestamp("2012-05-01")


class MeshOriginalTask(CTUEntityTask):
    entity_table = "mesh"
    target_col = "num"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class MondialOriginalTask(CTUEntityTask):
    entity_table = "target"
    target_col = "Target"
    task_type = TaskType.BINARY_CLASSIFICATION


class MooneyOriginalTask(CTULinkTask):
    entity_table = "uncle"
    task_type = TaskType.LINK_PREDICTION


class MovieLensOriginalTask(CTUEntityTask):
    entity_table = "users"
    target_col = "u_gender"
    task_type = TaskType.BINARY_CLASSIFICATION


class MuskSmallOriginalTask(CTUEntityTask):
    entity_table = "molecule"
    target_col = "class"
    task_type = TaskType.BINARY_CLASSIFICATION


class MuskLargeOriginalTask(CTUEntityTask):
    entity_table = "molecule"
    target_col = "class"
    task_type = TaskType.BINARY_CLASSIFICATION


class MutagenesisOriginalTask(CTUEntityTask):
    entity_table = "molecule"
    target_col = "mutagenic"
    task_type = TaskType.BINARY_CLASSIFICATION


class NCAAOriginalTask(CTUEntityTask):
    entity_table = "target"
    target_col = "team_id1_wins"
    task_type = TaskType.BINARY_CLASSIFICATION


class NCAATemporalTask(CTUEntityTaskTemporal):
    entity_table = "target"
    target_col = "team_id1_wins"
    task_type = TaskType.BINARY_CLASSIFICATION
    val_timestamp = pd.Timestamp("2010-11-01")
    test_timestamp = pd.Timestamp("2012-11-05")


class NorthwindOriginalTask(CTUEntityTask):
    entity_table = "Orders"
    target_col = "Freight"
    task_type = TaskType.REGRESSION


class NorthwindTemporalTask(CTUEntityTaskTemporal):
    entity_table = "Orders"
    target_col = "Freight"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("1998-02-01")
    test_timestamp = pd.Timestamp("1998-04-01")


class PimaOriginalTask(CTUEntityTask):
    entity_table = "pima"
    target_col = "arg2"
    task_type = TaskType.BINARY_CLASSIFICATION


class PremiereLeagueOriginalTask(CTUEntityTask):
    entity_table = "Matches"
    target_col = "ResultOfTeamHome"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class PremiereLeagueTemporalTask(CTUEntityTaskTemporal):
    entity_table = "Matches"
    target_col = "ResultOfTeamHome"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("2012-04-01")
    test_timestamp = pd.Timestamp("2012-05-01")


class RestbaseOriginalTask(CTUEntityTask):
    entity_table = "generalinfo"
    target_col = "review"
    task_type = TaskType.REGRESSION


class SakilaOriginalTask(CTUEntityTask):
    entity_table = "payment"
    target_col = "amount"
    task_type = TaskType.REGRESSION


class SakilaTemporalTask(CTUEntityTaskTemporal):
    entity_table = "payment"
    target_col = "amount"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2005-08-19")
    test_timestamp = pd.Timestamp("2005-08-21")


class SalesOriginalTask(CTUEntityTask):
    entity_table = "Sales"
    target_col = "Quantity"
    task_type = TaskType.REGRESSION


class SameGenOriginalTask(CTULinkTask):
    entity_table = "target"
    target_col = "target"
    task_type = TaskType.BINARY_CLASSIFICATION


class SAPOriginalTask(CTUEntityTask):
    entity_table = "Mailings"
    target_col = "RESPONSE"
    task_type = TaskType.BINARY_CLASSIFICATION


class SAPSalesTask(CTUEntityTask):
    entity_table = "Sales"
    target_col = "AMOUNT"
    task_type = TaskType.REGRESSION


class SAPSalesTemporalTask(CTUEntityTaskTemporal):
    entity_table = "Sales"
    target_col = "AMOUNT"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2007-06-10")
    test_timestamp = pd.Timestamp("2007-06-20")


class SatelliteOriginalTask(CTULinkTask):
    entity_table = "tm"
    link_table = "fault"
    task_type = TaskType.LINK_PREDICTION


class SeznamOriginalTask(CTUEntityTask):
    entity_table = "probehnuto"
    target_col = "kc_proklikano"
    task_type = TaskType.REGRESSION


class SeznamTemporalTask(CTUEntityTaskTemporal):
    entity_table = "probehnuto"
    target_col = "kc_proklikano"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2015-03-01")
    test_timestamp = pd.Timestamp("2015-07-01")


class SFScoresOriginalTask(CTUEntityTask):
    entity_table = "inspections"
    target_col = "score"
    task_type = TaskType.REGRESSION


class SFScoresTemporalTask(CTUEntityTaskTemporal):
    entity_table = "inspections"
    target_col = "score"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2016-03-01")
    test_timestamp = pd.Timestamp("2016-07-01")


class ShakespeareOriginalTask(CTULinkTask):
    entity_table = "paragraphs"
    link_table = "characters"
    task_type = TaskType.LINK_PREDICTION


class StatsOriginalTask(CTUEntityTask):
    entity_table = "users"
    target_col = "Reputation"
    task_type = TaskType.REGRESSION


class StatsTemporalTask(CTUEntityTaskTemporal):
    entity_table = "users"
    target_col = "Reputation"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2014-03-01")
    test_timestamp = pd.Timestamp("2014-06-01")


class StudentLoanOriginalTask(CTUEntityTask):
    entity_table = "no_payment_due"
    target_col = "bool"
    task_type = TaskType.BINARY_CLASSIFICATION


class ThrombosisOriginalTask(CTUEntityTask):
    entity_table = "Examination"
    target_col = "Thrombosis"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class ThrombosisTemporalTask(CTUEntityTaskTemporal):
    entity_table = "Examination"
    target_col = "Thrombosis"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("1996-01-01")
    test_timestamp = pd.Timestamp("1997-01-01")


class ToxicologyOriginalTask(CTUEntityTask):
    entity_table = "molecule"
    target_col = "label"
    task_type = TaskType.BINARY_CLASSIFICATION


class TPCCOriginalTask(CTUEntityTask):
    entity_table = "C_Customer"
    target_col = "c_credit"
    task_type = TaskType.BINARY_CLASSIFICATION


class TPCDOriginalTask(CTUEntityTask):
    entity_table = "dss_customer"
    target_col = "c_mktsegment"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class TPCDTemporalTask(CTUEntityTaskTemporal):
    entity_table = "dss_customer"
    target_col = "c_mktsegment"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("1997-01-01")
    test_timestamp = pd.Timestamp("1998-01-01")


class TPCDSOriginalTask(CTUEntityTask):
    entity_table = "customer"
    target_col = "c_preferred_cust_flag"
    task_type = TaskType.BINARY_CLASSIFICATION


class TPCDSTemporalTask(CTUEntityTaskTemporal):
    entity_table = "customer"
    target_col = "c_preferred_cust_flag"
    task_type = TaskType.BINARY_CLASSIFICATION
    val_timestamp = pd.Timestamp("2001-01-01")
    test_timestamp = pd.Timestamp("2002-01-01")


class TPCHOriginalTask(CTUEntityTask):
    entity_table = "customer"
    target_col = "c_acctbal"
    task_type = TaskType.REGRESSION


class TPCHTemporalTask(CTUEntityTaskTemporal):
    entity_table = "customer"
    target_col = "c_acctbal"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("1997-09-01")
    test_timestamp = pd.Timestamp("1998-03-01")


class TriazineOriginalTask(CTUEntityTask):
    entity_table = "molecule"
    target_col = "activity"
    task_type = TaskType.REGRESSION


class UWCSEOriginalTask(CTUEntityTask):
    entity_table = "person"
    target_col = "inPhase"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class VisualGenomeOriginalTask(CTUEntityTask):
    entity_table = "IMG_OBJ"
    target_col = "OBJ_CLASS"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class VOCOriginalTask(CTUEntityTask):
    entity_table = "voyages"
    target_col = "arrival_harbour"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class VOCTemporalTask(CTUEntityTaskTemporal):
    entity_table = "voyages"
    target_col = "arrival_harbour"
    task_type = TaskType.MULTICLASS_CLASSIFICATION
    val_timestamp = pd.Timestamp("1887-01-01")
    test_timestamp = pd.Timestamp("1902-01-01")


class WalmartOriginalTask(CTUEntityTask):
    entity_table = "train_table"
    target_col = "units"
    task_type = TaskType.REGRESSION


class WalmartTemporalTask(CTUEntityTaskTemporal):
    entity_table = "train_table"
    target_col = "units"
    task_type = TaskType.REGRESSION
    val_timestamp = pd.Timestamp("2014-01-01")
    test_timestamp = pd.Timestamp("2014-06-01")


class WebKPOriginalTask(CTUEntityTask):
    entity_table = "webpage"
    target_col = "class_label"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


class WorldOriginalTask(CTUEntityTask):
    entity_table = "Country"
    target_col = "Continent"
    task_type = TaskType.MULTICLASS_CLASSIFICATION


# fmt: off
__all__ = [
    "AccidentsOriginalTask", "AccidentsTemporalTask",
    "AdventureWorksOriginalTask", "AdventureWorksTemporalTask",
    "AirlineOriginalTask", "AirlineTemporalTask",
    "AtherosclerosisOriginalTask", 
    "BasketballMenOriginalTask",
    "BasketballWomenOriginalTask", 
    "BiodegradabilityOriginalTask",
    "BupaOriginalTask", 
    "CarcinogenesisOriginalTask", 
    "CDESchoolsOriginalTask", 
    "ChessOriginalTask", 
    "ClassicModelsOriginalTask", "ClassicModelsTemporalTask",
    "CORAOriginalTask", 
    "CountriesOriginalTask", 
    "CraftBeerOriginalTask", 
    "CreditOriginalTask", 
    "DallasOriginalTask", "DallasTemporalTask",
    "DCGOriginalTask", 
    "DiabetesOriginalTask",
    "DunurOriginalTask", 
    "EltiOriginalTask", 
    "EmployeeOriginalTask", "EmployeeTemporalTask",
    "ErgastF1OriginalTask", "ErgastF1TemporalTask",
    "ExpendituresOriginalTask", 
    "FinancialOriginalTask", "FinancialTemporalTask",
    "FNHKOriginalTask", "FNHKTemporalTask",
    "FTPOriginalTask", "FTPTemporalTask",
    "GeneeaOriginalTask", "GeneeaTemporalTask",
    "GenesOriginalTask", 
    "GOSalesOriginalTask", "GOSalesTemporalTask",
    "GrantsOriginalTask", "GrantsTemporalTask",
    "HepatitisOriginalTask", 
    "HockeyOriginalTask", "HockeyTemporalTask",
    "IMDbOriginalTask", "IMDbTemporalTask",
    "LahmanOriginalTask", "LahmanTemporalTask",
    "LegalActsOriginalTask", "LegalActsTemporalTask",
    "MeshOriginalTask", 
    "MondialOriginalTask", 
    "MooneyOriginalTask", 
    "MovieLensOriginalTask", 
    "MuskLargeOriginalTask", 
    "MuskSmallOriginalTask", 
    "MutagenesisOriginalTask",
    "NCAAOriginalTask", "NCAATemporalTask",
    "NorthwindOriginalTask", "NorthwindTemporalTask",
    "PimaOriginalTask", 
    "PremiereLeagueOriginalTask", "PremiereLeagueTemporalTask",
    "RestbaseOriginalTask",
    "SakilaOriginalTask", "SakilaTemporalTask",
    "SalesOriginalTask", 
    "SameGenOriginalTask", 
    "SAPOriginalTask", "SAPSalesTask", "SAPSalesTemporalTask",
    "SatelliteOriginalTask", 
    "SeznamOriginalTask", "SeznamTemporalTask",
    "SFScoresOriginalTask", "SFScoresTemporalTask",
    "ShakespeareOriginalTask", 
    "StatsOriginalTask", "StatsTemporalTask",
    "StudentLoanOriginalTask", 
    "ThrombosisOriginalTask", "ThrombosisTemporalTask",
    "ToxicologyOriginalTask",
    "TPCCOriginalTask", 
    "TPCDOriginalTask", "TPCDTemporalTask",
    "TPCDSOriginalTask", "TPCDSTemporalTask",
    "TPCHOriginalTask", "TPCHTemporalTask",
    "TriazineOriginalTask",
    "UWCSEOriginalTask", 
    "VisualGenomeOriginalTask", 
    "VOCOriginalTask", "VOCTemporalTask",
    "WalmartOriginalTask", "WalmartTemporalTask",
    "WebKPOriginalTask", 
    "WorldOriginalTask"
]
# fmt: on



from typing import Callable, Dict, List, Optional, Union
from numpy.typing import NDArray

from copy import deepcopy

import pandas as pd

from torch_frame import stype
from torch_frame.data import stats

from relbench.base import BaseTask, Database, Table, TaskType
from relbench.metrics import (
    accuracy,
    average_precision,
    f1,
    macro_f1,
    mae,
    micro_f1,
    mse,
    r2,
    roc_auc,
)

from ctu_relational.datasets import CTUDataset

__all__ = ["CTUBaseEntityTask"]


class CTUBaseEntityTask(BaseTask):
    timedelta = pd.Timedelta(-1)
    dataset: CTUDataset

    entity_col = "__PK__"
    entity_table: Optional[str] = None
    target_col: str

    _stats = None

    @property
    def metrics(self) -> List[Callable[[NDArray, NDArray], float]]:
        if self.task_type == TaskType.REGRESSION:
            return [mae, mse, r2]
        elif self.task_type == TaskType.BINARY_CLASSIFICATION:
            return [accuracy, average_precision, f1, roc_auc]

        elif self.task_type == TaskType.MULTICLASS_CLASSIFICATION:
            return [accuracy, macro_f1, micro_f1]

    @property
    def stats(self) -> Dict[stats.StatType, Union[list, tuple]]:
        if self._stats is None:
            db = self.dataset.get_db(upto_test_timestamp=False)
            table = self._get_full_table(db)
            target = table.df[self.target_col]
            self._set_stats(target)

        return self._stats

    def __init__(self, dataset: str, cache_dir: Optional[str] = None):
        ALLOWED_TASK_TYPES = [
            TaskType.REGRESSION,
            TaskType.BINARY_CLASSIFICATION,
            TaskType.MULTICLASS_CLASSIFICATION,
        ]
        if self.task_type not in ALLOWED_TASK_TYPES:
            raise ValueError(
                f"Task type {self.task_type.name} not allowed by CTUEntityTask."
            )

        super().__init__(dataset=dataset, cache_dir=cache_dir)

    def make_split(self, split: str, table: Table) -> Table:
        r"""Make a table using the task definition.

        Args:
            split: The split to be made.
            table: The table to be split.

        Returns:
            Table: The split table.

        To be implemented by subclass. The table rows need not be ordered
        deterministically.
        """

        raise NotImplementedError

    def get_sanitized_db(self, upto_test_timestamp: bool = True) -> Database:
        r"""Get the database object for the task without task target data.

        Args:
            upto_test_timestamp: If True, only return rows upto test_timestamp.
        Returns:
            Database: The database object.
        """

        _db = self.dataset.get_db(upto_test_timestamp=upto_test_timestamp)

        db = deepcopy(_db)

        db.table_dict[self.entity_table].df.drop(columns=[self.target_col], inplace=True)

        return db

    def _get_table(self, split: str) -> Table:
        r"""Helper function to get a table for a split."""

        db = self.dataset.get_db(upto_test_timestamp=False)

        full_table = self._get_full_table(db)

        if self.task_type in [
            TaskType.BINARY_CLASSIFICATION,
            TaskType.MULTICLASS_CLASSIFICATION,
        ]:
            full_table.df[self.target_col], unique = full_table.df[
                self.target_col
            ].factorize(sort=True)

        if self.task_type in [TaskType.BINARY_CLASSIFICATION, TaskType.REGRESSION]:
            full_table.df[self.target_col] = full_table.df[self.target_col].astype(float)
        else:
            full_table.df[self.target_col] = full_table.df[self.target_col].astype(int)

        split_table = self.make_split(split, full_table)

        return split_table

    def _get_full_table(self, db: Database) -> Table:
        time_col = db.table_dict[self.entity_table].time_col

        table = Table(
            df=db.table_dict[self.entity_table].df.copy(),
            fkey_col_to_pkey_table={self.entity_col: self.entity_table},
            pkey_col=None,
            time_col=time_col,
        )

        table.df = table.df[table.df[self.target_col].notna()]

        return table

    def _set_stats(self, target: pd.Series):
        t = {
            TaskType.REGRESSION: stype.numerical,
            TaskType.BINARY_CLASSIFICATION: stype.categorical,
            TaskType.MULTICLASS_CLASSIFICATION: stype.categorical,
            TaskType.MULTILABEL_CLASSIFICATION: stype.multicategorical,
        }[self.task_type]
        self._stats = {}
        for s in stats.StatType.stats_for_stype(t):
            self._stats[s] = s.compute(target)



import os

import pandas as pd

from relbench.datasets import register_dataset

from .db_dataset import DBDataset
from .ctu_dataset import CTUDataset

# fmt: off
from .ctu_datasets import (
    Accidents, AdventureWorks, Airline, Atherosclerosis, BasketballMen,
    BasketballWomen, Biodegradability, Bupa, Carcinogenesis,
    CDESchools, Chess, ClassicModels, CORA, Countries, CraftBeer, Credit,
    Dallas, DCG, Diabetes, Dunur, Elti, ErgastF1,
    Expenditures, Employee, Financial, FNHK, FTP, Geneea, Genes, GOSales,
    Grants, Hepatitis, Hockey, IMDb, Lahman, LegalActs, Mesh,
    Mondial, Mooney, MovieLens, MuskLarge, MuskSmall, Mutagenesis,
    NCAA, Northwind, Pima, PremiereLeague, Restbase, Sakila,
    Sales, SameGen, SAP, Satellite, Seznam, SFScores, Shakespeare, Stats,
    StudentLoan, Thrombosis, Toxicology, TPCC, TPCD, TPCDS, TPCH, Triazine,
    UWCSE, VisualGenome, VOC, Walmart, WebKP, World
)
# fmt: on


def get_dataset_info(dataset_name: str):
    info_df = pd.read_csv(os.path.join(os.path.dirname(__file__), "dataset-info.csv"))
    return info_df[info_df["dataset"] == dataset_name].iloc[0]


register_dataset("ctu-accidents", Accidents)
register_dataset("ctu-adventureworks", AdventureWorks)
register_dataset("ctu-airline", Airline)
register_dataset("ctu-atherosclerosis", Atherosclerosis)
register_dataset("ctu-basketballmen", BasketballMen)
register_dataset("ctu-basketballwomen", BasketballWomen)
register_dataset("ctu-biodegradability", Biodegradability)
register_dataset("ctu-bupa", Bupa)
register_dataset("ctu-carcinogenesis", Carcinogenesis)
register_dataset("ctu-cde", CDESchools)
register_dataset("ctu-chess", Chess)
register_dataset("ctu-classicmodels", ClassicModels)
register_dataset("ctu-cora", CORA)
register_dataset("ctu-countries", Countries)
register_dataset("ctu-craftbeer", CraftBeer)
register_dataset("ctu-credit", Credit)
register_dataset("ctu-dallas", Dallas)
register_dataset("ctu-dcg", DCG)
register_dataset("ctu-diabetes", Diabetes)
register_dataset("ctu-dunur", Dunur)
register_dataset("ctu-elti", Elti)
register_dataset("ctu-employee", Employee)
register_dataset("ctu-ergastf1", ErgastF1)
register_dataset("ctu-expenditures", Expenditures)
register_dataset("ctu-employee", Employee)
register_dataset("ctu-financial", Financial)
register_dataset("ctu-fnhk", FNHK)
register_dataset("ctu-ftp", FTP)
register_dataset("ctu-geneea", Geneea)
register_dataset("ctu-genes", Genes)
register_dataset("ctu-gosales", GOSales)
register_dataset("ctu-grants", Grants)
register_dataset("ctu-hepatitis", Hepatitis)
register_dataset("ctu-hockey", Hockey)
register_dataset("ctu-imdb", IMDb)
register_dataset("ctu-lahman", Lahman)
register_dataset("ctu-legalacts", LegalActs)
register_dataset("ctu-mesh", Mesh)
register_dataset("ctu-mondial", Mondial)
register_dataset("ctu-mooney", Mooney)
register_dataset("ctu-movielens", MovieLens)
register_dataset("ctu-musklarge", MuskLarge)
register_dataset("ctu-musksmall", MuskSmall)
register_dataset("ctu-mutagenesis", Mutagenesis)
register_dataset("ctu-ncaa", NCAA)
register_dataset("ctu-northwind", Northwind)
register_dataset("ctu-pima", Pima)
register_dataset("ctu-premiereleague", PremiereLeague)
register_dataset("ctu-restbase", Restbase)
register_dataset("ctu-sakila", Sakila)
register_dataset("ctu-sales", Sales)
register_dataset("ctu-samegen", SameGen)
register_dataset("ctu-sap", SAP)
register_dataset("ctu-satellite", Satellite)
register_dataset("ctu-seznam", Seznam)
register_dataset("ctu-sfscores", SFScores)
register_dataset("ctu-shakespeare", Shakespeare)
register_dataset("ctu-stats", Stats)
register_dataset("ctu-studentloan", StudentLoan)
register_dataset("ctu-thrombosis", Thrombosis)
register_dataset("ctu-toxicology", Toxicology)
register_dataset("ctu-tpcc", TPCC)
register_dataset("ctu-tpcd", TPCD)
register_dataset("ctu-tpcds", TPCDS)
register_dataset("ctu-tpch", TPCH)
register_dataset("ctu-triazine", Triazine)
register_dataset("ctu-uwcse", UWCSE)
register_dataset("ctu-visualgenome", VisualGenome)
register_dataset("ctu-voc", VOC)
register_dataset("ctu-walmart", Walmart)
register_dataset("ctu-webkp", WebKP)
register_dataset("ctu-world", World)

# fmt: off
__all__ = [
    "get_dataset_info",
    
    "DBDataset", "CTUDataset",
    
    "Accidents", "AdventureWorks", "Airline", "Atherosclerosis", "BasketballMen",
    "BasketballWomen", "Biodegradability", "Bupa", "Carcinogenesis",
    "CDESchools", "Chess", "ClassicModels", "CORA", "Countries", "CraftBeer", "Credit",
    "Dallas", "DCG", "Diabetes", "Dunur", "Elti", "Employee", "ErgastF1",
    "Expenditures", "Employee", "Financial", "FNHK", "FTP", "Geneea", "Genes", "GOSales",
    "Grants", "Hepatitis", "Hockey", "IMDb", "Lahman", "LegalActs", "Mesh",
    "Mondial", "Mooney", "MovieLens", "MuskLarge", "MuskSmall", "Mutagenesis",
    "NCAA", "Northwind", "Pima", "PremiereLeague", "Restbase", "Sakila",
    "Sales", "SameGen", "SAP", "Satellite", "Seznam", "SFScores", "Shakespeare", "Stats",
    "StudentLoan", "Thrombosis", "Toxicology", "TPCC", "TPCD", "TPCDS", "TPCH", "Triazine",
    "UWCSE", "VisualGenome", "VOC", "Walmart", "WebKP", "World"
]
# fmt: on



from typing import Callable, Dict, List, Optional
from numpy.typing import NDArray

import pandas as pd

from relbench.base import BaseTask, Database, Table


class CTULinkTask(BaseTask):
    timedelta = pd.Timedelta(-1)
    entity_col = "__PK__"
    entity_table: str
    target_col: str

    def make_table(self, db: Database, split: str) -> Table:
        raise NotImplementedError

    def get_sanitized_db(self, upto_test_timestamp: bool = True) -> Database:
        raise NotImplementedError

    def evaluate(
        self,
        pred: NDArray,
        target_table: Optional[Table] = None,
        metrics: Optional[List[Callable[[NDArray, NDArray], float]]] = None,
    ) -> Dict[str, float]:
        raise NotImplementedError

    def filter_dangling_entities(self, table: Table) -> Table:
        raise NotImplementedError



from typing import Callable, Dict, List, Optional
from numpy.typing import NDArray

import pandas as pd

from relbench.base import Table

from .ctu_entity_task_base import CTUBaseEntityTask

__all__ = ["CTUEntityTaskTemporal"]


class CTUEntityTaskTemporal(CTUBaseEntityTask):
    # To be set by subclass.
    val_timestamp: pd.Timestamp
    test_timestamp: pd.Timestamp

    def make_split(self, split: str, table: Table) -> Table:
        if table.time_col is None:
            raise ValueError("The table must have a time column.")

        table.df = table.df[table.df[table.time_col].notna()]

        if split == "train":
            table.df = table.df[table.df[table.time_col] < self.val_timestamp]
        elif split == "val":
            table.df = table.df[
                (table.df[table.time_col] >= self.val_timestamp)
                & (table.df[table.time_col] < self.test_timestamp)
            ]
        elif split == "test":
            table.df = table.df[table.df[table.time_col] >= self.test_timestamp]
        else:
            raise ValueError(f"Invalid split: {split}.")

        table.df = table.df[[self.entity_col, table.time_col, self.target_col]]

        return table

    def evaluate(
        self,
        pred: NDArray,
        target_table: Optional[Table] = None,
        metrics: Optional[List[Callable[[NDArray, NDArray], float]]] = None,
    ) -> Dict[str, float]:
        if metrics is None:
            metrics = self.metrics

        if target_table is None:
            target_table = self.get_table("test", mask_input_cols=False)

        target = target_table.df[self.target_col].to_numpy()
        if len(pred) != len(target):
            raise ValueError(
                f"The length of pred and target must be the same (got "
                f"{len(pred)} and {len(target)}, respectively)."
            )

        return {fn.__name__: fn(target, pred) for fn in metrics}



from .autodetect_schema import (
    check_predetermined_types,
    guess_column_stype,
    guess_table_stypes,
    guess_schema,
)
from .datetime import (
    standardize_db_dt,
    standardize_table_dt,
    convert_timedelta,
    TIMESTAMP_MAX,
    TIMESTAMP_MIN,
)
from .merge import merge_tf

__all__ = [
    "check_predetermined_types",
    "guess_column_stype",
    "guess_table_stypes",
    "guess_schema",
    "standardize_db_dt",
    "standardize_table_dt",
    "convert_timedelta",
    "TIMESTAMP_MAX",
    "TIMESTAMP_MIN",
    "merge_tf",
]



from relbench.tasks import register_task

from .ctu_entity_task_base import CTUBaseEntityTask
from .ctu_entity_task_classic import CTUEntityTask
from .ctu_entity_task_temporal import CTUEntityTaskTemporal

# fmt: off
from .ctu_tasks import (
    AccidentsOriginalTask, AccidentsTemporalTask, AdventureWorksOriginalTask, 
    AdventureWorksTemporalTask, AirlineOriginalTask, AirlineTemporalTask, 
    AtherosclerosisOriginalTask, BasketballMenOriginalTask, BasketballWomenOriginalTask,
    BiodegradabilityOriginalTask, BupaOriginalTask, CarcinogenesisOriginalTask,
    CDESchoolsOriginalTask, ChessOriginalTask, ClassicModelsOriginalTask,
    ClassicModelsTemporalTask, CORAOriginalTask, CountriesOriginalTask,
    CraftBeerOriginalTask, CreditOriginalTask, DallasOriginalTask, DallasTemporalTask,
    DCGOriginalTask, DiabetesOriginalTask, DunurOriginalTask, EltiOriginalTask,
    EmployeeOriginalTask, EmployeeTemporalTask, ErgastF1OriginalTask,
    ErgastF1TemporalTask, ExpendituresOriginalTask, FinancialOriginalTask,
    FinancialTemporalTask, FNHKOriginalTask, FNHKTemporalTask, FTPOriginalTask,
    FTPTemporalTask, GeneeaOriginalTask, GeneeaTemporalTask, GenesOriginalTask,
    GOSalesOriginalTask, GOSalesTemporalTask, GrantsOriginalTask, GrantsTemporalTask,
    HepatitisOriginalTask, HockeyOriginalTask, HockeyTemporalTask, IMDbOriginalTask,
    IMDbTemporalTask, LahmanOriginalTask, LahmanTemporalTask, LegalActsOriginalTask,
    LegalActsTemporalTask, MeshOriginalTask, MondialOriginalTask, MooneyOriginalTask,
    MovieLensOriginalTask, MuskLargeOriginalTask, MuskSmallOriginalTask,
    MutagenesisOriginalTask, NCAAOriginalTask, NCAATemporalTask, NorthwindOriginalTask,
    NorthwindTemporalTask, PimaOriginalTask, PremiereLeagueOriginalTask,
    PremiereLeagueTemporalTask, RestbaseOriginalTask, SakilaOriginalTask,
    SakilaTemporalTask, SalesOriginalTask, SameGenOriginalTask, SAPOriginalTask,
    SAPSalesTask, SAPSalesTemporalTask, SatelliteOriginalTask, SeznamOriginalTask,
    SeznamTemporalTask, SFScoresOriginalTask, SFScoresTemporalTask,
    ShakespeareOriginalTask, StatsOriginalTask, StatsTemporalTask, StudentLoanOriginalTask,
    ThrombosisOriginalTask, ThrombosisTemporalTask, ToxicologyOriginalTask,
    TPCCOriginalTask, TPCDOriginalTask, TPCDTemporalTask, TPCDSOriginalTask,
    TPCDSTemporalTask, TPCHOriginalTask, TPCHTemporalTask, TriazineOriginalTask,
    UWCSEOriginalTask, VisualGenomeOriginalTask, VOCOriginalTask, VOCTemporalTask,
    WalmartOriginalTask, WalmartTemporalTask, WebKPOriginalTask, WorldOriginalTask
)
# fmt: on

register_task("ctu-accidents", "accidents-original", AccidentsOriginalTask)

register_task("ctu-accidents", "accidents-temporal", AccidentsTemporalTask)

register_task("ctu-adventureworks", "adventureworks-original", AdventureWorksOriginalTask)
register_task("ctu-adventureworks", "adventureworks-temporal", AdventureWorksTemporalTask)

register_task("ctu-airline", "airline-original", AirlineOriginalTask)
register_task("ctu-airline", "airline-temporal", AirlineTemporalTask)

register_task(
    "ctu-atherosclerosis", "atherosclerosis-original", AtherosclerosisOriginalTask
)
register_task("ctu-basketballmen", "basketballmen-original", BasketballMenOriginalTask)
register_task(
    "ctu-basketballwomen", "basketballwomen-original", BasketballWomenOriginalTask
)
register_task(
    "ctu-biodegradability", "biodegradability-original", BiodegradabilityOriginalTask
)
register_task("ctu-bupa", "bupa-original", BupaOriginalTask)
register_task("ctu-carcinogenesis", "carcinogenesis-original", CarcinogenesisOriginalTask)
register_task("ctu-cde", "cde-original", CDESchoolsOriginalTask)
register_task("ctu-chess", "chess-original", ChessOriginalTask)

register_task("ctu-classicmodels", "classicmodels-original", ClassicModelsOriginalTask)
register_task("ctu-classicmodels", "classicmodels-temporal", ClassicModelsTemporalTask)

register_task("ctu-cora", "cora-original", CORAOriginalTask)
register_task("ctu-countries", "countries-original", CountriesOriginalTask)
register_task("ctu-craftbeer", "craftbeer-original", CraftBeerOriginalTask)
register_task("ctu-credit", "credit-original", CreditOriginalTask)

register_task("ctu-dallas", "dallas-original", DallasOriginalTask)
register_task("ctu-dallas", "dallas-temporal", DallasTemporalTask)

register_task("ctu-dcg", "dcg-original", DCGOriginalTask)
register_task("ctu-diabetes", "diabetes-original", DiabetesOriginalTask)
register_task("ctu-dunur", "dunur-original", DunurOriginalTask)
register_task("ctu-elti", "elti-original", EltiOriginalTask)

register_task("ctu-employee", "employee-original", EmployeeOriginalTask)
register_task("ctu-employee", "employee-temporal", EmployeeTemporalTask)

register_task("ctu-ergastf1", "ergastf1-original", ErgastF1OriginalTask)
# register_task("ctu-ergastf1", "ergastf1-temporal", ErgastF1TemporalTask)

register_task("ctu-expenditures", "expenditures-original", ExpendituresOriginalTask)

register_task("ctu-financial", "financial-original", FinancialOriginalTask)
register_task("ctu-financial", "financial-temporal", FinancialTemporalTask)

register_task("ctu-fnhk", "fnhk-original", FNHKOriginalTask)
register_task("ctu-fnhk", "fnhk-temporal", FNHKTemporalTask)

register_task("ctu-ftp", "ftp-original", FTPOriginalTask)
register_task("ctu-ftp", "ftp-temporal", FTPTemporalTask)

register_task("ctu-geneea", "geneea-original", GeneeaOriginalTask)
register_task("ctu-geneea", "geneea-temporal", GeneeaTemporalTask)

register_task("ctu-genes", "genes-original", GenesOriginalTask)

register_task("ctu-gosales", "gosales-original", GOSalesOriginalTask)
register_task("ctu-gosales", "gosales-temporal", GOSalesTemporalTask)

register_task("ctu-grants", "grants-original", GrantsOriginalTask)
register_task("ctu-grants", "grants-temporal", GrantsTemporalTask)

register_task("ctu-hepatitis", "hepatitis-original", HepatitisOriginalTask)

register_task("ctu-hockey", "hockey-original", HockeyOriginalTask)
# register_task("ctu-hockey", "hockey-temporal", HockeyTemporalTask)

register_task("ctu-imdb", "imdb-original", IMDbOriginalTask)
# register_task("ctu-imdb", "imdb-temporal", IMDbTemporalTask)

register_task("ctu-lahman", "lahman-original", LahmanOriginalTask)
register_task("ctu-lahman", "lahman-temporal", LahmanTemporalTask)

register_task("ctu-legalacts", "legalacts-original", LegalActsOriginalTask)
register_task("ctu-legalacts", "legalacts-temporal", LegalActsTemporalTask)

register_task("ctu-mesh", "mesh-original", MeshOriginalTask)
register_task("ctu-mondial", "mondial-original", MondialOriginalTask)
register_task("ctu-mooney", "mooney-original", MooneyOriginalTask)
register_task("ctu-movielens", "movielens-original", MovieLensOriginalTask)
register_task("ctu-musklarge", "musklarge-original", MuskLargeOriginalTask)
register_task("ctu-musksmall", "musksmall-original", MuskSmallOriginalTask)
register_task("ctu-mutagenesis", "mutagenesis-original", MutagenesisOriginalTask)

register_task("ctu-ncaa", "ncaa-original", NCAAOriginalTask)
# register_task("ctu-ncaa", "ncaa-temporal", NCAATemporalTask)

register_task("ctu-northwind", "northwind-original", NorthwindOriginalTask)
register_task("ctu-northwind", "northwind-temporal", NorthwindTemporalTask)

register_task("ctu-pima", "pima-original", PimaOriginalTask)

register_task("ctu-premiereleague", "premiereleague-original", PremiereLeagueOriginalTask)
register_task("ctu-premiereleague", "premiereleague-temporal", PremiereLeagueTemporalTask)

register_task("ctu-restbase", "restbase-original", RestbaseOriginalTask)

register_task("ctu-sakila", "sakila-original", SakilaOriginalTask)
register_task("ctu-sakila", "sakila-temporal", SakilaTemporalTask)

register_task("ctu-sales", "sales-original", SalesOriginalTask)
register_task("ctu-samegen", "samegen-original", SameGenOriginalTask)

register_task("ctu-sap", "sap-original", SAPOriginalTask)
register_task("ctu-sap", "sap-sales", SAPSalesTask)
register_task("ctu-sap", "sap-sales-temporal", SAPSalesTemporalTask)

register_task("ctu-satellite", "satellite-original", SatelliteOriginalTask)

register_task("ctu-seznam", "seznam-original", SeznamOriginalTask)
register_task("ctu-seznam", "seznam-temporal", SeznamTemporalTask)

register_task("ctu-sfscores", "sfscores-original", SFScoresOriginalTask)
register_task("ctu-sfscores", "sfscores-temporal", SFScoresTemporalTask)

register_task("ctu-shakespeare", "shakespeare-original", ShakespeareOriginalTask)

register_task("ctu-stats", "stats-original", StatsOriginalTask)
register_task("ctu-stats", "stats-temporal", StatsTemporalTask)

register_task("ctu-studentloan", "studentloan-original", StudentLoanOriginalTask)

register_task("ctu-thrombosis", "thrombosis-original", ThrombosisOriginalTask)
# register_task("ctu-thrombosis", "thrombosis-temporal", ThrombosisTemporalTask)

register_task("ctu-toxicology", "toxicology-original", ToxicologyOriginalTask)
register_task("ctu-tpcc", "tpcc-original", TPCCOriginalTask)

register_task("ctu-tpcd", "tpcd-original", TPCDOriginalTask)
# register_task("ctu-tpcd", "tpcd-temporal", TPCDTemporalTask)

register_task("ctu-tpcds", "tpcds-original", TPCDSOriginalTask)
register_task("ctu-tpcds", "tpcds-temporal", TPCDSTemporalTask)

register_task("ctu-tpch", "tpch-original", TPCHOriginalTask)
# register_task("ctu-tpch", "tpch-temporal", TPCHTemporalTask)

register_task("ctu-triazine", "triazine-original", TriazineOriginalTask)
register_task("ctu-uwcse", "uwcse-original", UWCSEOriginalTask)
register_task("ctu-visualgenome", "visualgenome-original", VisualGenomeOriginalTask)

register_task("ctu-voc", "voc-original", VOCOriginalTask)
register_task("ctu-voc", "voc-temporal", VOCTemporalTask)

register_task("ctu-walmart", "walmart-original", WalmartOriginalTask)
register_task("ctu-walmart", "walmart-temporal", WalmartTemporalTask)

register_task("ctu-webkp", "webkp-original", WebKPOriginalTask)
register_task("ctu-world", "world-original", WorldOriginalTask)


# fmt: off
__all__ = [
    "CTUBaseEntityTask", "CTUEntityTask", "CTUEntityTaskTemporal",
    
    "AccidentsOriginalTask", "AccidentsTemporalTask", "AdventureWorksOriginalTask", 
    "AdventureWorksTemporalTask", "AirlineOriginalTask", "AirlineTemporalTask", 
    "AtherosclerosisOriginalTask", "BasketballMenOriginalTask", "BasketballWomenOriginalTask",
    "BiodegradabilityOriginalTask", "BupaOriginalTask", "CarcinogenesisOriginalTask",
    "CDESchoolsOriginalTask", "ChessOriginalTask", "ClassicModelsOriginalTask",
    "ClassicModelsTemporalTask", "CORAOriginalTask", "CountriesOriginalTask",
    "CraftBeerOriginalTask", "CreditOriginalTask", "DallasOriginalTask", "DallasTemporalTask",
    "DCGOriginalTask", "DiabetesOriginalTask", "DunurOriginalTask", "EltiOriginalTask",
    "EmployeeOriginalTask", "EmployeeTemporalTask", "ErgastF1OriginalTask", "ErgastF1TemporalTask",
    "ExpendituresOriginalTask", "FinancialOriginalTask", "FinancialTemporalTask", "FNHKOriginalTask",
    "FNHKTemporalTask", "FTPOriginalTask", "FTPTemporalTask", "GeneeaOriginalTask",
    "GeneeaTemporalTask", "GenesOriginalTask", "GOSalesOriginalTask", "GOSalesTemporalTask",
    "GrantsOriginalTask", "GrantsTemporalTask", "HepatitisOriginalTask", "HockeyOriginalTask",
    "HockeyTemporalTask", "IMDbOriginalTask", "IMDbTemporalTask", "LahmanOriginalTask",
    "LahmanTemporalTask", "LegalActsOriginalTask", "LegalActsTemporalTask", "MeshOriginalTask",
    "MondialOriginalTask", "MooneyOriginalTask", "MovieLensOriginalTask", "MuskLargeOriginalTask",
    "MuskSmallOriginalTask", "MutagenesisOriginalTask", "NCAAOriginalTask", "NCAATemporalTask",
    "NorthwindOriginalTask", "NorthwindTemporalTask", "PimaOriginalTask", "PremiereLeagueOriginalTask",
    "PremiereLeagueTemporalTask", "RestbaseOriginalTask", "SakilaOriginalTask", "SakilaTemporalTask",
    "SalesOriginalTask", "SameGenOriginalTask", "SAPOriginalTask", "SAPSalesTask", "SAPSalesTemporalTask",
    "SatelliteOriginalTask", "SeznamOriginalTask", "SeznamTemporalTask", "SFScoresOriginalTask",
    "SFScoresTemporalTask", "ShakespeareOriginalTask", "StatsOriginalTask", "StatsTemporalTask",
    "StudentLoanOriginalTask", "ThrombosisOriginalTask", "ThrombosisTemporalTask",
    "ToxicologyOriginalTask", "TPCCOriginalTask", "TPCDOriginalTask", "TPCDTemporalTask",
    "TPCDSOriginalTask", "TPCDSTemporalTask", "TPCHOriginalTask", "TPCHTemporalTask",
    "TriazineOriginalTask", "UWCSEOriginalTask", "VisualGenomeOriginalTask", "VOCOriginalTask",
    "VOCTemporalTask", "WalmartOriginalTask", "WalmartTemporalTask", "WebKPOriginalTask", "WorldOriginalTask"
]
# fmt: on



from typing import Literal, Optional

from collections import defaultdict
import copy

import torch

from torch_frame import stype, TensorFrame
from torch_frame.data import MultiEmbeddingTensor
from torch_frame.typing import TensorData


def merge_tf(
    left_tf: TensorFrame,
    right_tf: TensorFrame,
    left_idx: torch.Tensor,
    right_idx: torch.Tensor,
    left_prefix: Optional[str] = None,
    right_prefix: Optional[str] = None,
    how: Literal["left"] = "left",
) -> TensorFrame:
    """Merge two tensor frames."""

    left_col_names_dict = defaultdict(list)
    left_col_names_dict.update(
        {
            st: [f"{left_prefix}{c}" for c in left_tf.col_names_dict[st]]
            for st in left_tf.stypes
        }
        if left_prefix is not None
        else copy.deepcopy(left_tf.col_names_dict)
    )

    right_col_names_dict = defaultdict(list)
    right_col_names_dict.update(
        {
            st: [f"{right_prefix}{c}" for c in right_tf.col_names_dict[st]]
            for st in right_tf.stypes
        }
        if right_prefix is not None
        else copy.deepcopy(right_tf.col_names_dict)
    )

    left_cols = []
    for cols in left_col_names_dict.values():
        left_cols.extend(cols)

    right_cols = []
    for cols in right_col_names_dict.values():
        right_cols.extend(cols)

    if len(set(left_cols).intersection(set(right_cols))) > 0:
        raise ValueError(f"Column names overlap. Left: {left_cols} Right: {right_cols}")

    col_names_dict = {
        st: left_col_names_dict[st] + right_col_names_dict[st]
        for st in left_tf.stypes + right_tf.stypes
    }

    if how != "left":
        raise NotImplementedError("Only left join is supported.")

    num_rows = left_tf.num_rows
    feat_dict = copy.deepcopy(left_tf.feat_dict)
    for st in right_tf.stypes:
        new_data = _init_stype_features(
            st,
            len(right_col_names_dict[st]),
            num_rows,
            sparse_data=right_tf.feat_dict[st][right_idx],
            idx=left_idx,
        )

        st_data = feat_dict.get(st, None)
        feat_dict[st] = (
            new_data if st_data is None else _concat_features(st, [st_data, new_data])
        )
    return TensorFrame(
        feat_dict=feat_dict, col_names_dict=col_names_dict, y=left_tf.y, num_rows=num_rows
    )


def _init_stype_features(
    s: stype, num_cols: int, num_rows: int, sparse_data: TensorData, idx: torch.Tensor
) -> TensorData:
    data: TensorData = None
    if s == stype.numerical:
        data = torch.full((num_rows, num_cols), torch.nan)
        data[idx] = sparse_data
    elif s == stype.categorical:
        data = torch.full((num_rows, num_cols), -1)
        data[idx] = sparse_data
    elif s == stype.timestamp:
        data = torch.full((num_rows, num_cols, 7), -1)
        data[idx] = sparse_data

    elif s.use_multi_embedding_tensor:
        values = torch.full((num_rows, sparse_data.values.shape[1]), -1.0)
        values[idx] = sparse_data.values
        data = MultiEmbeddingTensor(
            num_rows=num_rows,
            num_cols=num_cols,
            values=values,
            offset=sparse_data.offset,
        )
    else:
        raise ValueError(f"Unknown stype: {s}")

    return data


def _concat_features(s: stype, data_list: list[TensorData]) -> TensorData:
    if s == stype.numerical or s == stype.categorical or s == stype.timestamp:
        return torch.cat(data_list, dim=1)
    elif s.use_multi_embedding_tensor:
        return MultiEmbeddingTensor.cat(data_list, dim=1)
    else:
        raise ValueError(f"Unknown stype: {s}")


__all__ = ["merge_tf"]
